//Receives a file from the FileServer over IPv6. Use ::1 for the loopback
//address rather than 127.0.0.1

#include "Dasyne.hpp"
#include <iostream>

using namespace std;
using namespace libdasyne;

//The port number to communicate over
const uint16_t PORT_NUMBER = 7890;

//False if we encountered an error and should stop the program
static bool runProgram = true;

//Track the time it takes to transfer the file
static DTime transferTime;

#ifdef DA_WINDOWS_VERSION
    #include "WindowsExamplesTools.hpp"
#endif

//_______________ Event Handlers _______________

//Called when we connect to the server
void connectHandler(std::string netMessage,
                    NetworkCode socketStatus)
{
    if (socketStatus == DA_SUCCESS)
    {
        cout << "Connected to server." << endl;
        transferTime.updateTime();
    }
    else
    {
        cout << netMessage << endl;
        cout << "Was it an IPv6 address?" << endl;
        runProgram = false;
    }
}//connectHandler

//Called when the server sends us data
void receiveHandler(DBuffer &theData)
{
    DField *oneField = NULL;//the current field to extract from fileProtocol
    DProtocol fileProtocol;//parses file information from the received data
    DBufferIter dIter;//iterates through fileData
    DBuffer fileBytes;//contains the complete file
    string fileName;//the name of the original file
    uint32_t serverChecksum = 0;//the file checksum, computed by the server
    uint32_t myChecksum = 0;//the file checksum, computed from the received data
    ofstream outfile;//the file to write to

    cout << "Time to receive file: " << transferTime.timeElapsed() << " milliseconds" << endl;

    //### Define the protocol we expect to receive ###
    fileProtocol.appendField(new DField("checksum", DA_UINT32_T));
    fileProtocol.appendField(new DField("file contents", DA_DBUFFER));
    fileProtocol.appendField(new DField("file name", DA_STRING));

    //### Split the message into distinct fields ###
    if (!fileProtocol.parseMessage(theData))
    {
        cout << fileProtocol.getErrMsg() << endl;
        runProgram = false;
    }

    //### Compare checksums ###
    if (runProgram)
    {
        oneField = fileProtocol.getField("file contents");
        oneField->getDBuffer(0, fileBytes);

        oneField = fileProtocol.getField("checksum");
        oneField->getValue(0, serverChecksum);

        oneField = fileProtocol.getField("file name");
        oneField->getString(0, fileName);

        myChecksum = fileBytes.crc32();
        if (myChecksum == serverChecksum)
            cout << "Checksum valid!" << endl;
        else
        {
            cout << "Got a corrupted file!" << endl;
            cout << "Server's checksum: " << serverChecksum << endl;
            cout << "My checksum: " << myChecksum << endl;
            runProgram = false;
        }
    }

    //### Save the file to disk ###
    if (runProgram)
    {
        fileName.append(".dwnld");
        cout << "Writing to disk as " << fileName << "..." << endl;
        outfile.open(fileName.c_str(), ios::binary);
        if (!outfile.is_open())
            cout << "Couldn't create file!" << endl;
        else
        {
            dIter.setBuffer(&fileBytes);

            while (!dIter.finished())
            {
                outfile << dIter.value();
                dIter.next();
            }

            outfile.close();
        }
    }

    runProgram = false;
}//receiveHandler

//Called when an error occurs. If socketOK is false, the connection failed
void errorHandler(string errMsg,
                  bool socketOK)
{
    cout << errMsg << endl;
    if (!socketOK)
        runProgram = false;
}//errorHandler

//_______________ Main Function _______________

int main(int argc,
         char *argv[])
{
    DEngine netEngine;//handles network events
    DDispatcher *client = new DDispatcher(&netEngine);//connects to the server and receives the file
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Windows specific configuration - see WindowsExamplesTools.hpp ###
    #ifdef DA_WINDOWS_VERSION
        if (!initializeWindows())
            runProgram = false;
    #endif

    //### Set up event handlers and connect to the server ###
    if ((argc == 2) && (runProgram))
    {
        client->addHandler(&receiveHandler);
        client->setErrorHandler(&errorHandler);
        client->startConnect(argv[1], PORT_NUMBER, DA_IPV6, &connectHandler);
    }
    else
    {
        cout << "Usage: fileclient <ipv6 hostname>" << endl;
        cout << "Eg: fileclient ::1" << endl;
        runProgram = false;
    }

    while (runProgram)
    {
        //### Dispatch any queued events ###
        eventsRun = netEngine.update();
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg() << endl;
            runProgram = false;
        }

        //### Sleep if we have nothing to do ###
        else if (eventsRun == 0)
            DEngine::sleep(1);
    }

    //### Delete the classes that use netEngine before deleting netEngine ###
    delete client;

    return 0;
}//main
