/* Sends TCP and UDP messages to the server at specified intervals,
   and measures the latency of the server's response.
*/

#include "Dasyne.hpp"
#include <iostream>
#include <math.h>

using namespace std;
using namespace libdasyne;

//_______________ Function Prototypes _______________

bool extractParameters(int argc,
                       char *argv[]);
void startSending(void);
void showUsage(void);
void showResults(void);

//_______________ Static Variables _______________

static vector<DTime*> sendTimes;//the time when each packet was sent
static vector<DTime*> receiveTimes;//the time when that packet was received

static DEngine netEngine;//generates network events
static DDispatcher *client = new DDispatcher(&netEngine);//handles all communication to the server
static DUdpModule *udpModule = new DUdpModule(&netEngine);//provides UDP support to the DDispatcher
static DTimer *sendTimer = new DTimer(&netEngine);//controls when messages are sent
static DTimer *graceTimer = new DTimer(&netEngine);//waits for any stray packets to arrive
static DBuffer theMessage;//the message to send to the server

static string serverName;//the server's hostname
static uint16_t messageNumber = 0;//count the messages as we send them
static int payloadSize = 0;//the size of the payload in bytes
static unsigned int totalPackets = 0;//the number of packets to send with each protocol
static const uint16_t SERVER_UDP_PORT = 5978;//The UDP port to send to
static const uint16_t SERVER_TCP_PORT = 5979;//The TCP port to connect to
static const unsigned int SEND_TIME = 5000;//the number of milliseconds of sending for each protocol
static const unsigned int GRACE_TIME = 3000;//the number of millisecondss to wait before outputting results

static bool runProgram = true;//false if we should abort the program
static bool sendingTcp = true;//false if we're currently sending UDP

#ifdef DA_WINDOWS_VERSION
    #include "WindowsExamplesTools.hpp"
#endif

//_______________ Event Handlers _______________

//Called when we connect to the server
void connectHandler(std::string netMessage,
                    NetworkCode socketStatus)
{
    if (socketStatus == DA_SUCCESS)
        cout << "Connected to " << client->getRemoteAddress() << endl;
    else
    {
        cout << netMessage << endl;
        runProgram = false;
    }
}//connectHandler

//Called when we receive an 'enable UDP' request from the server
void udpRequestHandler(uint16_t requestedPort,
                       DUdpModule*& dispatcherUDP)
{
    if (udpModule->bind(DA_IPV4, requestedPort))
    {
        cout << "Bound to UDP port " << requestedPort << endl;
        dispatcherUDP = udpModule;
        startSending();
    }
    else
    {
        cout << udpModule->getErrMsg();
        runProgram = false;
    }
}//udpRequestHandler

//Start sending messages to the server
void startSending(void)
{
    cout << endl;
    if (sendingTcp)
        cout << "Sending TCP..." << endl;
    else
        cout << "Sending UDP..." << endl;

    for (unsigned int i = 0; i < sendTimes.size(); i++)
        delete sendTimes.at(i);
    sendTimes.clear();

    for (unsigned int i = 0; i < receiveTimes.size(); i++)
        delete receiveTimes.at(i);
    receiveTimes.clear();

    sendTimes.resize(totalPackets, NULL);
    receiveTimes.resize(totalPackets, NULL);

    messageNumber = 0;
    sendTimer->start();
}//startSending

//Called periodically to send data
void sendMessage(void)
{
    DBuffer numberBytes;//the bytes in messageNumber

    numberBytes.appendVariable(messageNumber);

    theMessage.setByte(0, numberBytes.getByte(0));
    theMessage.setByte(1, numberBytes.getByte(1));

    sendTimes.at(messageNumber) = new DTime;
    sendTimes.at(messageNumber)->updateTime();

    if (sendingTcp)
    {
        theMessage.setByte(2, 0);
        client->sendTcp(theMessage);
    }
    else
    {
        theMessage.setByte(2, 1);
        client->sendUdp(theMessage);
    }

    messageNumber++;
    if (messageNumber >= totalPackets)
    {
        sendTimer->stop();
        graceTimer->start();
    }
}

//Called when the server mirrors our data back to us
void receiveHandler(DBuffer &theData)
{
    DTime *currentTime = NULL;//store the current time
    uint16_t messageNumber = 0;//the identifier for the received message
    bool haveCorrectProtocol = true;//false if we receive a TCP message when we're expecting UDP

    //### Verify that we received the entire payload ###
    if (theData.size() != static_cast<unsigned int>(payloadSize))
    {
        cout << "Expected a payload of " << payloadSize << " bytes, got " << theData.size() << " bytes" << endl;
        runProgram = false;
    }

    //### Read the message number ###
    if (runProgram)
    {
        theData.extractVariable(0, messageNumber);

        if (messageNumber >= receiveTimes.size())
        {
            cout << "Received invalid message number " << messageNumber << endl;
            runProgram  = false;
        }
    }

    //### Ignore TCP messages when we're measuring UDP latency ###
    if (runProgram)
    {
        if ((theData.getByte(2) == 0) && (!sendingTcp))
            haveCorrectProtocol = false;
    }

    //### Store the received time ###
    if ((runProgram) && (haveCorrectProtocol))
    {
        if (receiveTimes.at(messageNumber) == NULL)
        {
            currentTime = new DTime();
            currentTime->updateTime();
            receiveTimes.at(messageNumber) = currentTime;
        }
        else
        {
            cout << "Received duplicate message number " << messageNumber << endl;
            runProgram = false;
        }
    }
}//receiveHandler

void graceTimeElapsed(void)
{
    graceTimer->stop();
    showResults();

    //### Calculate UDP latency ###
    if (sendingTcp)
    {
        sendingTcp = false;
        startSending();
    }

    //### We're finished ###
    else
        runProgram = false;
}

void showResults(void)
{
    DTime *currentSendTime = NULL;
    DTime *currentRecvTime = NULL;
    int totalMeasurements = 0;//the number of send/recv pairs
    int elapsedTime = 0;//the difference between the send and receive times, in milliseconds
    int maxLatency = 0;//the maximum latency we encountered, in ms
    int minLatency = 10000;//the minimum latency we encountered, in ms
    int totalElapsed = 0;//the total amount of latency, from all received messages

    for (unsigned int i = 0; i < sendTimes.size(); i++)
    {
        currentSendTime = sendTimes.at(i);
        currentRecvTime = receiveTimes.at(i);

        if (currentRecvTime != NULL)
        {
            elapsedTime = (currentRecvTime->getSeconds() - currentSendTime->getSeconds()) * 1000;
            elapsedTime += currentRecvTime->getMilliseconds() - currentSendTime->getMilliseconds();

            if (elapsedTime > maxLatency)
                maxLatency = elapsedTime;
            if (elapsedTime < minLatency)
                minLatency = elapsedTime;

            totalElapsed += elapsedTime;
            totalMeasurements++;
        }
    }

    cout << endl;
    if (sendingTcp)
        cout << "TCP Results" << endl;
    else
        cout << "UDP Results" << endl;

    cout << "    Messages sent: " << sendTimes.size() << endl;
    cout << "    Messages received: " << totalMeasurements << endl;

    if (totalMeasurements > 0)
    {
        cout << "    Min Latency: " << minLatency << " ms" << endl;
        cout << "    Max Latency: " << maxLatency << " ms" << endl;

        if (totalElapsed != 0)
            cout << "    Avg Latency: " << static_cast<double>(totalElapsed) / static_cast<double>(totalMeasurements) << " ms" << endl;
    }
}//showResults

//Called when an error occurs. If socketOK is false, the connection failed
void errorHandler(string errMsg,
                  bool socketOK)
{
    cout << errMsg << endl;
    if (!socketOK)
        runProgram = false;
}//errorHandler

//_______________ Main Function _______________

int main(int argc,
         char *argv[])
{
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Windows specific configuration - see WindowsExamplesTools.hpp ###
    #ifdef DA_WINDOWS_VERSION
        if (!initializeWindows())
            runProgram = false;
    #endif

    //### Extract command-line parameters ###
    if (runProgram)
    {
        if (!extractParameters(argc, argv))
            runProgram = false;
    }

    //### Set up event handlers and connect to the server ###
    if (runProgram)
    {
        graceTimer->setHandler(&graceTimeElapsed);
        graceTimer->setInterval(GRACE_TIME);

        udpModule->setErrorHandler(errorHandler);

        client->addHandler(&receiveHandler);
        client->setErrorHandler(&errorHandler);
        client->setUdpRequestHandler(&udpRequestHandler);
        client->startConnect(argv[1], SERVER_TCP_PORT, DA_IPV4, &connectHandler);
    }

    while (runProgram)
    {
        //### Dispatch any queued events ###
        eventsRun = netEngine.update();
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg() << endl;
            runProgram = false;
        }

        //### Sleep if we have nothing to do ###
        #ifndef DA_WINDOWS_VERSION
            else if (eventsRun == 0)
                DEngine::sleep(1);
        #endif
    }

    //### Free memory ###
    for (unsigned int i = 0; i < sendTimes.size(); i++)
        delete sendTimes.at(i);
    sendTimes.clear();

    for (unsigned int i = 0; i < receiveTimes.size(); i++)
        delete receiveTimes.at(i);
    receiveTimes.clear();

    delete udpModule;
    delete client;
    delete sendTimer;
    delete graceTimer;

    return 0;
}//main

bool extractParameters(int argc,
                       char *argv[])
{
    double packetsToSend = 0;//the number of packets we can send within SEND_TIME
    int sendInterval = 0;//the interval at which to send messages, in milliseconds
    bool result = true;//false on malformed parameters

    //### Verify that the proper number of arguments were given ###
    if (argc != 4)
    {
        showUsage();
        result = false;
    }

    //### Extract the server's address ###
    if (result)
        serverName = argv[1];

    //### Extract the send interval in milliseconds ###
    if (result)
    {
        if (!DEngine::strToInt(argv[2], sendInterval))
        {
            cout << "Expected a send interval, got " << argv[2] << endl;
            showUsage();
            result = false;
        }
    }

    //### Verify that the send interval is valid ###
    if (result)
    {
        if (sendInterval > 0)
        {
            sendTimer->setInterval(sendInterval);
            sendTimer->setHandler(sendMessage);

            packetsToSend = static_cast<double>(SEND_TIME) / static_cast<double>(sendInterval);
            totalPackets = static_cast<unsigned int>(ceil(packetsToSend));
        }
        else
        {
            cout << "Got a send interval of " << sendInterval << " ms" << endl;
            showUsage();
            result = false;
        }
    }

    //### Extract the payload size ###
    if (result)
    {
        if (!DEngine::strToInt(argv[3], payloadSize))
        {
            cout << "Expected a payload size, got " << argv[3] << endl;
            showUsage();
            result = false;
        }
    }

    //### Verify that the payload size is valid ###
    if (result)
    {
        if (payloadSize > 0)
        {
            if (payloadSize < 3)//make room for the packet number + send type
                payloadSize = 3;
        }
        else
        {
            cout << "Payload size must be positive, got " << payloadSize << endl;
            showUsage();
            result = false;
        }
    }

    //### Generate the payload ###
    if (result)
    {
        for (int i = 0; i < payloadSize; i++)
            theMessage.appendByte(i % 255);
    }

    return result;
}//extractParameters

void showUsage(void)
{
    cout << "Usage: latclient <server hostname> <interval (ms)> <payload (bytes)>" << endl;
    cout << endl;
}//showUsage
