/* Mirrors TCP and UDP messages back to the sender, for determining
   the latency of network messages with different sizes.
*/

#include "Dasyne.hpp"
#include <iostream>

using namespace std;
using namespace libdasyne;

const uint16_t CLIENT_UDP_PORT = 5977;//The UDP port number to bind to
const uint16_t SERVER_UDP_PORT = 5978;//The UDP port number to bind to
const uint16_t SERVER_TCP_PORT = 5979;//The TCP port number to bind to
static bool runProgram = true; //false if we should stop the server
static DEngine netEngine;//calls event handlers for networking classes
static DUdpModule *udpModule = new DUdpModule(&netEngine);//handles UDP data for one or more DDispatchers
DDispatcher *theClient = NULL;//the client to send packets to, or NULL if none are connected

#ifdef DA_WINDOWS_VERSION
    #include "WindowsExamplesTools.hpp"
#endif

//_______________ DClient Event Handlers _______________

//Event handler called when we've finished setting up UDP communication
void udpReady(UdpStatus udpStatus)
{
    if (udpStatus == DA_UDP_OK)
        cout << "UDP available!" << endl;
    else
    {
        if (udpStatus == DA_UDP_REFUSED)
            cout << "The remote host refused or was unable to activate UDP" << endl;
        else if (udpStatus == DA_UDP_TIMEOUT)
            cout << "The remote host did not respond to our UDP request" << endl;

        netEngine.deleteAfter(theClient);
        theClient = NULL;
    }
}

//Called when the client sends us a message
void messageReceived(DBuffer &message)
{
    if (message.size() < 3)
    {
        cout << "Received an incomplete message!" << endl;
        runProgram = false;
    }

    //Mirror TCP messages
    if (message.getByte(2) == 0)
        theClient->sendTcp(message);

    //Mirror UDP messages
    else if (message.getByte(2) == 1)
        theClient->sendUdp(message);

    //Unknown message type!
    else
    {
        cout << "Received an unknown protocol: " << static_cast<int>(message.getByte(2)) << endl;
        netEngine.deleteAfter(theClient);
        theClient = NULL;
    }
}//messageReceived

//Event handler called when a TCP network error occurs.
void tcpErrorHandler(string message,
                     bool socketOK)
{
    if (!socketOK)
    {
        cout << "TCP Error: " << message << endl;
        netEngine.deleteAfter(theClient);
        theClient = NULL;
    }
    else
        cout << message << endl;
}//tcpErrorHandler

//_______________ DUdpModule Event Handlers _______________

//Event handler called when a UDP network error occurs.
void udpErrorHandler(string message,
                     bool socketOK)
{
    if (!socketOK)
    {
        cout << "UDP Error: " << message << endl;
        runProgram = false;
    }
    else
        cout << message << endl;
}//udpErrorHandler

//_______________ DServer Event Handlers _______________

//Event handler called when there's an error accepting TCP connections
void listenErrorHandler(string message)
{
    cout << "TCP Listen Error: " << message << endl;
    runProgram = false;
}

//Called when a client connects
void connectHandler(boost::asio::ip::tcp::socket *newSocket)
{
    if (theClient == NULL)
    {
        theClient = new DDispatcher(&netEngine, newSocket);
        theClient->addHandler(&messageReceived);
        theClient->setErrorHandler(&tcpErrorHandler);

        cout << endl;
        cout << "A client connected from " << theClient->getRemoteAddress() << endl;
        theClient->requestUdp(udpModule, CLIENT_UDP_PORT, &udpReady);
    }
    else
        cout << "A client tried to connect, but another is already connected." << endl;
}//connectHandler

//_______________ Main Function _______________

int main(int argc,
         char *argv[])
{
    DDispatcher clientConnection(&netEngine);//sends and receives data
    DServer tcpListener(&netEngine);//listens for TCP connections
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Windows specific configuration - see WindowsExamplesTools.hpp ###
    #ifdef DA_WINDOWS_VERSION
        if (!initializeWindows())
            runProgram = false;
    #endif

    //### Bind the UDP socket ###
    if (runProgram)
    {
        udpModule->setErrorHandler(&udpErrorHandler);
        if (udpModule->bind(libdasyne::DA_IPV4, SERVER_UDP_PORT))
            cout << "Listening on UDP port " << SERVER_UDP_PORT << endl;
        else
        {
            runProgram = false;
            cout << "Couldn't bind: " << udpModule->getErrMsg() << endl;
        }
    }

    //### Bind the TCP socket ###
    if (runProgram)
    {
        tcpListener.setErrorHandler(&listenErrorHandler);
        if (tcpListener.bind(DA_IPV4, SERVER_TCP_PORT))
            cout << "Listening on TCP port " << SERVER_TCP_PORT << endl;
        else
        {
            runProgram = false;
            cout << "Couldn't bind: " << tcpListener.getErrMsg() << endl;
        }
    }

    //### Listen for TCP connections ###
    if (runProgram)
    {
        if (tcpListener.startListening(&connectHandler))
        {
            cout << endl;
            cout << "Mirroring messages back to sender, hit Ctrl-C to exit" << endl;
        }
        else
        {
            runProgram = false;
            cout << "Couldn't listen for TCP connections: " << tcpListener.getErrMsg() << endl;
        }
    }

    //### Periodically call event handlers ###
    while (runProgram)
    {
        eventsRun = netEngine.update();
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg();
            runProgram = false;
        }

        //Sleep if we have nothing to do
        else if (eventsRun == 0)
            DEngine::sleep(1);
    }

    //### Free memory ###
    delete udpModule;

    return 0;
}//main
