/* A generic TCP proxy server. Data from incoming transmissions is
   routed to the specified address and port, and vice-versa.
*/

#include "Dasyne.hpp"
#include <iostream>

using namespace std;
using namespace libdasyne;
class Client;
class Server;

static bool runProgram = true;//false if we should stop the server
static Client* theClient = NULL;//the connection to the client
static Server* theServer = NULL;//the connection to the server

static DEngine netEngine;//calls event handlers for networking classes
static DServer *myServer = new DServer(&netEngine);//listens for client connections

//The address and port of the server to connect to
static string destAddress;
static uint16_t destPort = 0;

//Only accept incoming connections from this address
static string whitelistIP;

//True if we've connected to the server and are ready to act as a proxy.
static bool serverConnected = false;

//The client may send us data before we've connected to the server. Store it here
//and send it once we connect.
static DBuffer sendBuffer;

#ifdef DA_WINDOWS_VERSION
    #include "WindowsExamplesTools.hpp"
#endif

//Display a message about how to use this utility.
void showUsage();

//Convert a string to a 2-byte port number, returns false on failure.
bool extractPort(string textPort,
                 uint16_t &port);

//Event handlers called when we receive data from the client or server
void receivedServerData(DBuffer &theData);
void receivedClientData(DBuffer &theData);

//Event handler called when there's a TCP error
void connectionError(string message,
                     bool socketOK);

class Server : public DClient
{
    public:
        Server(DEngine *netEngine) : DClient(netEngine)
        {
            setReceiveHandler(&receivedServerData);
            setErrorHandler(&connectionError);

            startConnect(destAddress, destPort, DA_IPV4, boost::bind(&Server::connectHandler, this, _1, _2));
        }//constructor

    private:
        void connectHandler(std::string netMessage,
                            NetworkCode socketStatus)
        {
            if (socketStatus == DA_SUCCESS)
            {
                cout << "Connected to " << getRemoteAddress() << endl;
                serverConnected = true;

                if (sendBuffer.size() > 0)
                {
                    send(sendBuffer);
                    sendBuffer.clear();
                }
            }
            else
                connectionError(netMessage, false);
        }//connectHandler
};//Server

class Client : public DClient
{
    public:
        //constructor, passes the DEngine and an existing connection to DClient
        Client(boost::asio::ip::tcp::socket *newSocket) : DClient(&netEngine, newSocket)
        {
            sendBuffer.clear();
            setReceiveHandler(&receivedClientData);
            setErrorHandler(&connectionError);
        }//constructor
};//Client

void receivedServerData(DBuffer &theData)
{
    if (theClient != NULL)
        theClient->send(theData);
}

void receivedClientData(DBuffer &theData)
{
    if (!serverConnected)
        sendBuffer.append(theData);
    else if (theServer != NULL)
        theServer->send(theData);
}

void connectionError(string message,
                     bool socketOK)
{
    if (!socketOK)
    {
        cout << message << endl;
        cout << "Disconnected." << endl;
        cout << endl;

        netEngine.deleteAfter(theServer);
        netEngine.deleteAfter(theClient);
        theServer = NULL;
        theClient = NULL;
        serverConnected = false;
        sendBuffer.clear();
    }
    else
        cout << message << endl;
}

//_______________ DServer Event Handlers _______________

//Called when a client connects
void connectHandler(boost::asio::ip::tcp::socket *newSocket)
{
    bool allOK = true;//false if we couldn't accept this connection

    //### Verify that no-one else is connected ###
    if (theClient != NULL)
    {
        cout << "A client tried to connect from " << theClient->getRemoteAddress()
             << ", but another client is already connected" << endl;
        allOK = false;
    }

    //### Check if the client is allowed to connect ###
    if (allOK)
    {
        theClient = new Client(newSocket);

        cout << "A client connected from " << theClient->getRemoteAddress() << endl;
        if ((whitelistIP.size() > 0) && (theClient->getRemoteAddress() != whitelistIP))
        {
            cout << "But we're only allowing connections from " << whitelistIP
                << ". Disconnecting!" << endl;

            netEngine.deleteSocket(theClient);
            theClient = NULL;
            allOK = false;
        }
    }

    //### Connect to the server ###
    if (allOK)
    {
        if (theServer != NULL)
            delete theServer;

        theServer = new Server(&netEngine);
        serverConnected = false;
    }
}//connectHandler

//Event handler, called when there's an error accepting connections
void errorHandler(string errMsg)
{
    cout << errMsg << endl;
    runProgram = false;
}//errorHandler

//_______________ Main Function _______________

int main(int argc,
         char *argv[])
{
    string userText;//text entered by the user
    int eventsRun = 0;//the number of events dispatched by a call to update()
    uint16_t listenPort = 0;//the port the server will listen on

    //### Verify that the correct number of arguments were specified ###
    if ((argc < 4) || (argc > 5))
    {
        showUsage();
        runProgram = false;
    }

    //### Windows specific configuration - see WindowsExamplesTools.hpp ###
    #ifdef DA_WINDOWS_VERSION
        if (runProgram)
        {
            if (!initializeWindows())
                runProgram = false;
        }
    #endif

    //### Extract the port to listen on ###
    if (runProgram)
    {
        if (!extractPort(argv[1], listenPort))
            runProgram = false;
    }

    //### Extract the destination port and address ###
    if (runProgram)
    {
        destAddress = argv[2];
        if (!extractPort(argv[3], destPort))
            runProgram = false;
    }

    //### Bind the listen socket ###
    if (runProgram)
    {
        if (!myServer->bind(libdasyne::DA_IPV4, listenPort))
        {
            runProgram = false;
            cout << "Couldn't bind: " << myServer->getErrMsg() << endl;
        }
    }

    //### Start listening for connections ###
    if (runProgram)
    {
        if (argc == 5)
            whitelistIP = argv[4];

        myServer->setErrorHandler(&errorHandler);
        if (myServer->startListening(&connectHandler))
            cout << "Listening for TCP connections on port " << listenPort << ", hit Ctrl-C to exit." << endl;
        else
        {
            runProgram = false;
            cout << myServer->getErrMsg() << endl;
        }
    }

    //### Periodically call event handlers ###
    while (runProgram)
    {
        eventsRun = netEngine.update();
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg();
            runProgram = false;
        }

        //Sleep if we have nothing to do
        else if (eventsRun == 0)
            DEngine::sleep(1);
    }

    //### Delete the objects that use DEngine before it's deleted ###
    delete theClient;
    delete theServer;

    return 0;
}//main

bool extractPort(string textPort,
                 uint16_t &port)
{
    int intPort = 0;
    bool result = true;//false on malformed textPort

    if (!DEngine::strToInt(textPort, intPort))
    {
        cout << "Invalid port number: " << textPort << endl;
        showUsage();
        result = false;
    }

    if (result)
    {
        if ((intPort >= 0) && (intPort <= 65535))
            port = intPort;
        else
        {
            cout << "Invalid port number " << intPort << endl;
            showUsage();
            result = false;
        }
    }

    return result;
}

void showUsage()
{
    cout << "Usage:" << endl;
    cout << "tcpproxy <listen_port> <destination_IP> <destination_port> [allowed_IP]" << endl;
    cout << endl;
    cout << "    listen_port       The port the proxy server will listen on." << endl;
    cout << "    destination_IP    The address the proxy server will connect to." << endl;
    cout << "    destination_port  The port the proxy server will connect to." << endl;
    cout << "    allowed_IP        For security, only allow incoming connections from this address." << endl;
    cout << endl;
    cout << endl;
    cout << " ____________________" << endl;
    cout << "|    User Program    |" << endl;
    cout << "|                    |" << endl;
    cout << "| allowed_IP         |_______________" << endl;
    cout << "|____________________|               |" << endl;
    cout << "                                     |" << endl;
    cout << " ____________________________     ___V________________" << endl;
    cout << "|    Destination Computer    |   |    Proxy Server    |" << endl;
    cout << "|                            |<--|                    |" << endl;
    cout << "| destination_IP             |   | listen_port        |" << endl;
    cout << "| destination_port           |   |____________________|" << endl;
    cout << "|____________________________|" << endl;
}//showUsage
