/* Tests the throughput of the link to a tserver program. See the
  ThroughputServer.cpp file for details.
*/

#include "Dasyne.hpp"
#include <iostream>

using namespace std;
using namespace libdasyne;

//_______________ Static Variables _______________

static DEngine netEngine;//handles network events
static DClient *client = NULL;//handles our connection to the server
const uint16_t PORT_NUMBER = 5127;//The port number to communicate over
static double kbReceived = 0;//The total number of kilobytes received
static int receiveBufferSize = 0;//The number of kilobytes to allocate in our receive buffer
static bool runProgram = true;//False if we encountered an error and should stop the program

#ifdef DA_WINDOWS_VERSION
    #include "WindowsExamplesTools.hpp"
#endif

//Shows tclient's command-line usage
void showUsage(void);

//_______________ Event Handlers _______________

//Called when the connect operation finishes
void connectHandler(string message,
                    libdasyne::NetworkCode status)
{
    if (status == libdasyne::DA_SUCCESS)
    {
        cout << "Connected to " << client->getRemoteAddress() << endl;
        cout << "My address: " << client->getLocalAddress() << endl;
    }
    else
    {
        cout << message << endl;
        runProgram = false;
    }
}//connectHandler

//Event handler, called when we receive data
void receiveHandler(DBuffer &receivedBytes)
{
    kbReceived += receivedBytes.size() / 1024.0;
}//receiveHandler

//Event handler, called when a network error occurs
void errorHandler(string message,
                  bool socketOK)
{
    cout << message << endl;

    if (!socketOK)
    {
        cout << "Connection lost." << endl;
        runProgram = false;
    }
}//errorHandler

//_______________ Main Function _______________

int main(int argc,
         char *argv[])
{
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Verify that the proper number of arguments were given ###
    if (argc != 3)
    {
        showUsage();
        runProgram = false;
    }

    //### Windows specific configuration - see WindowsExamplesTools.hpp ###
    #ifdef DA_WINDOWS_VERSION
        if (runProgram)
        {
            if (!initializeWindows())
                runProgram = false;
        }
    #endif

    //### Extract the buffer size parameter ###
    if (runProgram)
    {
        if (!DEngine::strToInt(argv[2], receiveBufferSize))
        {
            showUsage();
            runProgram = false;
        }
    }

    //### Create the client and connect ###
    if (runProgram)
    {
        if (receiveBufferSize > 0)
        {
            client = new DClient(&netEngine, receiveBufferSize * 1024);
            client->setErrorHandler(&errorHandler);
            client->setReceiveHandler(&receiveHandler);
            client->startConnect(argv[1], PORT_NUMBER, DA_IPV4, &connectHandler);
        }
        else
        {
            showUsage();
            runProgram = false;
        }
    }

    while (runProgram)
    {
        //### Dispatch any queued events ###
        eventsRun = netEngine.update();
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg() << endl;
            runProgram = false;
        }

        //### Sleep if we have nothing to do ###
        #ifndef DA_WINDOWS_VERSION
            else if (eventsRun == 0)
                DEngine::sleep(1);
        #endif
    }

    //### Output results ###
    if (kbReceived > 1048576)
        cout << "Received " << kbReceived / 1048576.0 << " GB" << endl;
    else if (kbReceived > 1048576)
        cout << "Received " << kbReceived / 1024.0 << " MB" << endl;
    else
        cout << "Received " << kbReceived << " KB" << endl;

    //### Delete the classes that use netEngine before netEngine is deleted ###
    delete client;

    return 0;
}//main

void showUsage(void)
{
    cout << "Usage: tclient <host_name> <buffer_KB>" << endl;
    cout << endl;
}//showUsage
