/* Tests the maximum throughput available to one or more client programs.
   We create a buffer of random data, and send it as often as possible to all clients
   for 15 seconds. Then, the program calculates the total bandwidth. The buffer size is
   adjustable as a command-line parameter, and we can use multiple clients to test the
   effects of multiple connections over one pipe.

   Meant to be used with the tclient program.
*/

#include "Dasyne.hpp"
#include <iostream>

using namespace std;
using namespace libdasyne;
class Client;

//_______________ Static Variables _______________

static const uint16_t PORT_NUMBER = 5127;//The port number to communicate over
static const int TEST_TIME = 15;//the number of seconds to run the test

static vector<Client*> theClients;//a list of clients who've connected
static DEngine netEngine;//calls event handlers for networking classes
DServer *myServer = new DServer(&netEngine);//listens for client connections
static DTime testTime;//stop the test after TEST_TIME seconds
static DBuffer theData;//the data to send

static double totalSent = 0;//the total number of kilobytes sent
static int numClients = 0;//the number of clients that should connect before the test starts
static int sendBufferSize = 0;//the size of the send buffers to use, in kilobytes

static bool runProgram = true; //false if we should stop the server
static bool runningTest = false;//true if the test is currently running
static DTimer *msgTimer = NULL;//periodically display our progress

#ifdef DA_WINDOWS_VERSION
    #include "WindowsExamplesTools.hpp"
#endif

//Shows tserver's command-line usage
void showUsage(void);

//Show the amount of data sent and the average send rate
void showResults(void);

//_______________ Client Class _______________

class Client : public DClient
{
    public:
        //constructor, passes the network engine and an existing connection to DClient
        Client(DEngine *newNetEngine,
               boost::asio::ip::tcp::socket *newSocket) : DClient(newNetEngine, newSocket, sendBufferSize)
        {
            setSendHandler(boost::bind(&Client::sendHandler, this));
            setErrorHandler(boost::bind(&Client::errorHandler, this, _1, _2));
        }//constructor

    private:
        //Called when we've sent all data
        void sendHandler(void)
        {
            totalSent += theData.size() / 1024.0;
            send(theData);
        }//sendHandler

        //Event handler called when a network error occurs
        void errorHandler(std::string message,
                          bool networkOK)
        {
            if (!networkOK)
            {
                cout << "Client disconnected, stopping test." << endl;
                runProgram = false;
            }

            cout << "Error: " << message << endl;
        }//errorHandler
};//Client

//_______________ DServer Event Handlers _______________

//Called when a client connects
void connectHandler(boost::asio::ip::tcp::socket *newSocket)
{
    Client *newClient = new Client(&netEngine, newSocket);

    cout << "A client connected from " << newClient->getRemoteAddress() << endl;
    theClients.push_back(newClient);

    if ((int)theClients.size() < numClients)
        cout << "Waiting for " << numClients - theClients.size() << " more." << endl;
    else
    {
        cout << "Running test..." << endl;
        myServer->close();

        runningTest = true;
        testTime.updateTime();
        msgTimer->start();

        for (unsigned int i = 0; i < theClients.size(); i++)
            theClients.at(i)->send(theData);
    }

    cout << endl;
}//connectHandler

//Called when there's an error accepting connections
void errorHandler(string errMsg)
{
    cout << "Error accepting connections: " << errMsg << endl;
    runProgram = false;
}//errorHandler

//_______________ Main Function _______________

int main(int argc,
         char *argv[])
{
    int eventsRun = 0;//the number of events dispatched by a call to update()
    bool finished = false;//true if the test completed

    //### Verify that the parameters were given ###
    if (argc != 3)
    {
        showUsage();
        runProgram = false;
    }

    //### Extract the number of clients parameter ###
    if (runProgram)
    {
        if (!DEngine::strToInt(argv[1], numClients))
        {
            showUsage();
            runProgram = false;
        }
    }

    //### Verify that the number of clients is valid ###
    if (runProgram)
    {
        if (numClients <= 0)
        {
            showUsage();
            runProgram = false;
        }
    }

    //### Extract the buffer size parameter ###
    if (runProgram)
    {
        if (!DEngine::strToInt(argv[1], sendBufferSize))
        {
            showUsage();
            runProgram = false;
        }
    }

    //### Verify that the buffer size is valid ###
    if (runProgram)
    {
        if (sendBufferSize > 0)
        {
            if (sendBufferSize >= 1024)
                cout << "Woah - send buffer is over one megabyte! Continuing..." << endl;

            sendBufferSize *= 1024;

            for (int i = 0; i < sendBufferSize; i++)
                theData.appendByte(i % 255);
        }
        else
        {
            showUsage();
            runProgram = false;
        }
    }

    //### Windows specific configuration - see WindowsExamplesTools.hpp ###
    #ifdef DA_WINDOWS_VERSION
        if (runProgram)
        {
            if (!initializeWindows())
                runProgram = false;
        }
    #endif

    //### Bind the listen socket ###
    if (runProgram)
    {
        if (!myServer->bind(libdasyne::DA_IPV4, PORT_NUMBER))
        {
            runProgram = false;
            cout << "Couldn't bind: " << myServer->getErrMsg() << endl;
        }
    }

    //### Start listening for connections ###
    if (runProgram)
    {
        myServer->setErrorHandler(&errorHandler);
        if (myServer->startListening(&connectHandler))
            cout << "Listening for TCP connections..." << endl;
        else
        {
            runProgram = false;
            cout << "Couldn't listen: " << myServer->getErrMsg() << endl;
        }
    }

    //### Periodically display the amount transferred ###
    if (runProgram)
    {
        msgTimer = new DTimer(&netEngine);
        msgTimer->setInterval(3000);
        msgTimer->setHandler(&showResults);
    }

    //### Periodically call event handlers ###
    while ((runProgram) && (!finished))
    {
        //Dispatch events
        eventsRun = netEngine.update();
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg();
            runProgram = false;
        }

        //Sleep if we have nothing to do, Windows machines won't schedule the process
        //frequently enough to make this a valid test if we sleep here.
        #ifndef DA_WINDOWS_VERSION
            else if (eventsRun == 0)
                DEngine::sleep(1);
        #endif

        //Check when the test time elapses
        if (runningTest)
        {
            if (testTime.timeElapsed() > TEST_TIME * 1000)
                finished = true;
        }
    }

    //### Output the test results ###
    if (runProgram)
    {
        cout << endl;
        cout << "Final Results: " << endl;
        showResults();
    }

    //### Delete the objects that use DEngine before it's deleted ###
    for (unsigned int i = 0; i < theClients.size(); i++)
        delete theClients.at(i);
    theClients.clear();

    delete msgTimer;
    delete myServer;

    return 0;
}//main

void showResults(void)
{
    float elapsedTime = (float)testTime.timeElapsed() / 1000.0;
    float sendRate = 0;//the maximum rate at which we can send, in bytes/second

    if (totalSent > 1048576)
        cout << "Sent " << totalSent / 1048576.0 << " GB in ";
    else if (totalSent > 1024)
        cout << "Sent " << totalSent / 1024.0 << " MB in ";
    else
        cout << "Sent " << totalSent << " KB in ";
    cout << elapsedTime << " seconds" << endl;

    sendRate = (float)totalSent / elapsedTime;
    if (sendRate > 1048576)
        cout << "Send rate " << sendRate / 1048576.0 << " GB/second" << endl;
    else if (sendRate > 1024)
        cout << "Send rate " << sendRate / 1024.0 << " MB/second" << endl;
    else
        cout << "Send rate " << sendRate << " KB/second" << endl;

    cout << endl;
}//showResults

void showUsage(void)
{
    cout << "    Usage: tserver <number_of_clients> <buffer_KB>" << endl;
    cout << endl;
    cout << "    tserver will wait until the specified number of clients have connected" << endl;
    cout << "    before starting the test." << endl;
}//showUsage
