#include "Dasyne.hpp"
#include <iostream>

using namespace std;
using namespace libdasyne;

//_______________ Function Prototypes _______________

bool extractParameters(int argc,
                       char *argv[]);
void showResults(void);
void showUsage(void);

//_______________ Static Variables _______________

static const uint16_t PORT_NUMBER = 5978;//The port number to send to
static const unsigned int SEND_TIME = 5000;//the number of milliseconds to run the program

static DEngine netEngine;//handles network events
static DUdpSocket *udpSocket = NULL;//sends and receives UDP packets
static DTimer *dataTimer = new DTimer(&netEngine);//tells us when to send the next datagram
static DTimer *graceTimer = new DTimer(&netEngine);//wait after we've finished sending to receive all datagrams
static DBuffer payload;//the payload to send to the server
static vector<DTime*> sendTimes;//the time at which we sent each packet
static vector<DTime*> receiveTimes;//the time at which each packet was received
static string serverName;//the server's hostname
static string serverAddress;//the server's IP address

static unsigned int timeElapsed = 0;//the number of milliseconds that have elapsed so far
static int localPort = 0;//the local port number we'll bind to
static int sendInterval = 0;//the interval at which to send packets, in milliseconds
static int payloadSize = 0;//the size of the payload in bytes

static bool runProgram = true;//False if we encountered an error and should stop the program

#ifdef DA_WINDOWS_VERSION
    #include "WindowsExamplesTools.hpp"
#endif

//_______________ Event Handlers _______________

//Event handler to periodically send packets to the server
void timerExpired(void)
{
    DTime *currentTime = NULL;//store the time when each datagram was sent
    DBuffer packetLabel;//number each packet as we send it

    timeElapsed += sendInterval;
    if (timeElapsed > SEND_TIME)
    {
        dataTimer->stop();
        graceTimer->setInterval(3000);
        graceTimer->setHandler(&showResults);
        graceTimer->start();
    }//if timeElapsed
    else
    {
        currentTime = new DTime;
        currentTime->useEngine(&netEngine);
        currentTime->updateTime();

        packetLabel.appendVariable(static_cast<uint16_t>(sendTimes.size()));
        payload.setByte(payload.size() - 2, packetLabel.getByte(0));
        payload.setByte(payload.size() - 1, packetLabel.getByte(1));

        sendTimes.push_back(currentTime);
        receiveTimes.push_back(NULL);

        udpSocket->send(payload, serverAddress, PORT_NUMBER);
    }//else timeElapsed
}//timerExpired

//Handler called when we've resolved the server's hostname
void resolveHandler(vector<string> addressList)
{
    //### Obtain an IP address from addressList ###
    if (addressList.size() > 0)
    {
        cout << "Addresses: ";
        for (unsigned int i = 0; i < addressList.size(); i++)
            cout << addressList.at(i) << " ";
        cout << endl;

        serverAddress = addressList.at(0);
        cout << "Using: " << serverAddress << endl;
    }//if size()

    else
    {
        cout << udpSocket->getErrMsg() << endl;
        runProgram = false;
    }//else size()

    //### Start sending packets to the server ###
    if (runProgram)
    {
        //Leave a byte for the packet number
        for (int i = 0; i < payloadSize; i++)
            payload.appendByte(i % 255);

        dataTimer->setInterval(sendInterval);
        dataTimer->setHandler(&timerExpired);
        dataTimer->start();
        cout << "Sending datagrams..." << endl;
    }//if runProgram
}//resolveHandler

//Event handler, called when we receive data
void receiveHandler(DBuffer &receivedData,
                    string &sourceAddress,
                    uint16_t sourcePort)
{
    DTime *currentTime = NULL;//the current timee
    DBuffer packetLabel;//the packet label we set in timerExpired()
    uint16_t packetNumber = 0;//int equivalent of packetLabel
    int receivedSize = receivedData.size();
    bool allOK = true;//false if we received an invalid datagram

    //### Verify that we got a packet with the proper payload size ###
    if (receivedSize != payloadSize)
    {
        cout << "Got a packet with a payload of " << receivedSize << " bytes!?" << endl;
        allOK = false;
    }//if receivedSize

    //### Extract the packet number ###
    if (allOK)
    {
        packetLabel.appendByte(receivedData.getByte(receivedSize - 2));
        packetLabel.appendByte(receivedData.getByte(receivedSize - 1));
        packetLabel.extractVariable(0, packetNumber);

        if (packetNumber >= receiveTimes.size())
        {
            cout << "Got datagram number " << packetNumber
                 << ", but we've only sent " << receiveTimes.size() << " datagrams!" << endl;
            allOK = false;
        }//if packetNumber
    }//if allOK

    //### Store the current time ###
    if (allOK)
    {
        if (receiveTimes.at(packetNumber) == NULL)
        {
            currentTime = new DTime;
            currentTime->updateTime();
            receiveTimes.at(packetNumber) = currentTime;
        }//if receiveTimes
        else
        {
            cout << "Received duplicate datagram " << packetNumber << endl;
            allOK = false;
        }//else receiveTimes
    }//if allOK
}//receiveHandler

//Event handler, called when a network error occurs
void errorHandler(string message,
                  bool socketOK)
{
    cout << message << endl;
    if (!socketOK)
        runProgram = false;
}//errorHandler

//Event handler, called when we've finished waiting for mirrored packets
void showResults(void)
{
    DTime *oneTime = NULL;//the current packet's receive time
    unsigned int packetsReceived = 0;//the total number of packets we received
    unsigned int maxLatency = 0;//the maximum latency from all received packets
    unsigned int minLatency = 10000;//the minimum latency from all received packets
    unsigned int currentLatency = 0;//the latency of the current packet
    float totalLatency = 0;//the sum of the latencies of all packets

    for (unsigned int i = 0; i < receiveTimes.size(); i++)
    {
        oneTime = receiveTimes.at(i);

        if (oneTime != NULL)
        {
            currentLatency = receiveTimes.at(i)->getSeconds() - sendTimes.at(i)->getSeconds();
            currentLatency *= 1000;
            currentLatency += receiveTimes.at(i)->getMilliseconds() - sendTimes.at(i)->getMilliseconds();

            if (currentLatency < minLatency)
                minLatency = currentLatency;
            if (currentLatency > maxLatency)
                maxLatency = currentLatency;
            totalLatency += currentLatency;

            packetsReceived++;
        }//if oneTime
    }//for i

    if (packetsReceived == 0)
        cout << "No packets were received from the server!" << endl;
    else
    {
        cout << endl;
        cout << "Datagrams sent: " << sendTimes.size() << endl;
        cout << "Datagrams received: " << packetsReceived << endl;
        cout << "Min Latency: " << minLatency << " ms" << endl;
        cout << "Max Latency: " << maxLatency << " ms" << endl;
        cout << "Average Latency: "
            << static_cast<float>(totalLatency) / static_cast<float>(packetsReceived)
            << " ms " << endl;
        cout << endl;
    }

    runProgram = false;
}//showResults

//_______________ Main Function _______________

int main(int argc,
         char *argv[])
{
    int eventsRun = 0;//the number of events dispatched by a call to update()

    //### Extract command-line parameters ###
    if (!extractParameters(argc, argv))
        runProgram = false;

    //### Windows specific configuration - see WindowsExamplesTools.hpp ###
    #ifdef DA_WINDOWS_VERSION
        if (runProgram)
        {
            if (!initializeWindows())
                runProgram = false;
        }
    #endif

    //### Bind the socket ###
    if (runProgram)
    {
        udpSocket = new DUdpSocket(&netEngine);

        udpSocket->setErrorHandler(&errorHandler);
        udpSocket->setReceiveHandler(&receiveHandler);
        if (udpSocket->bind(DA_IPV4, localPort))
            cout << "UDP bound to port " << localPort << endl;
        else
        {
            cout << udpSocket->getErrMsg() << endl;
            runProgram = false;
        }
    }

    //### Perform hostname resolution ###
    if (runProgram)
    {
        cout << "Resolving hostname " << serverName << endl;
        udpSocket->resolveHost(serverName, DA_IPV4, &resolveHandler);
    }

    while (runProgram)
    {
        //### Dispatch any queued events ###
        eventsRun = netEngine.update();
        if (eventsRun < 0)
        {
            cout << netEngine.getErrMsg() << endl;
            runProgram = false;
        }

        //### Sleep if we have nothing to do ###
        #ifndef DA_WINDOWS_VERSION
            else if (eventsRun == 0)
                DEngine::sleep(1);
        #endif
    }

    //### Delete the classes that use netEngine before netEngine is deleted ###
    for (unsigned int i = 0; i < sendTimes.size(); i++)
        delete sendTimes.at(i);

    for (unsigned int i = 0; i < receiveTimes.size(); i++)
        delete receiveTimes.at(i);

    delete udpSocket;
    delete dataTimer;
    delete graceTimer;

    return 0;
}//main

bool extractParameters(int argc,
                       char *argv[])
{
    bool result = true;//false on malformed parameters

    //### Verify that the proper number of arguments were given ###
    if (argc != 5)
    {
        showUsage();
        result = false;
    }//if argc

    //### Extract the server address and our port number ###
    if (result)
    {
        serverName = argv[1];
        if (!DEngine::strToInt(argv[2], localPort))
        {
            cout << "Expected a port number, got " << argv[2] << endl;
            showUsage();
            result = false;
        }//if !strToInt()
    }//if result

    //### Verify that the port number is valid ###
    if (result)
    {
        if ((localPort < 0) || (localPort > 65535))
        {
            cout << "Port number should be between 0 and 65535, got " << localPort << endl;
            showUsage();
            result = false;
        }//if localPort || localPort
    }//if result

    //### Extract the send interval in milliseconds ###
    if (result)
    {
        if (!DEngine::strToInt(argv[3], sendInterval))
        {
            cout << "Expected a send interval, got " << argv[3] << endl;
            showUsage();
            result = false;
        }//if !strToInt()
    }//if result

    //### Verify that the send interval is valid ###
    if (result)
    {
        if (sendInterval <= 0)
        {
            cout << "Got a send interval of " << sendInterval << " ms" << endl;
            showUsage();
            result = false;
        }//if sendInterval
    }//if result

    //### Extract the payload size ###
    if (result)
    {
        if (!DEngine::strToInt(argv[4], payloadSize))
        {
            cout << "Expected a payload size, got " << argv[4] << endl;
            showUsage();
            result = false;
        }//if !strToInt()
    }//if result

    //### Verify that the payload size is valid ###
    if (result)
    {
        if (payloadSize <= 0)
        {
            cout << "Payload size must be 2 bytes or larger, got " << payloadSize << endl;
            showUsage();
            result = false;
        }//if payloadSize
    }//if result

    return result;
}//extractParameters

void showUsage(void)
{
    cout << "Usage: udpclient <server hostname> <my UDP port number> <interval (ms)> <payload (bytes)>" << endl;
    cout << endl;
}//showUsage
