/*
 * Copyright (C) 2005-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DBytes.hpp"
#include "DSafeArray.hpp"

using namespace std;
using namespace libdasyne;

#ifdef DA_DEBUG_MODE
    long int DBytes::numInstances = 0;
#endif

DBytes::DBytes(void)
{
    #ifdef DA_DEBUG_MODE
        numInstances++;
        arraysAdded = 0;
    #endif

    startFilledBytes = 0;
    endFilledBytes = 0;

    clear();
}//constructor

DBytes::DBytes(const DBytes &source)
{
    #ifdef DA_DEBUG_MODE
        numInstances++;
        arraysAdded = 0;
    #endif

    startFilledBytes = 0;
    endFilledBytes = 0;
    clear();

    append(source);
}//copy constructor

DBytes::~DBytes(void)
{
    #ifdef DA_DEBUG_MODE
        numInstances--;
        arraysAdded -= byteData.size();
    #else
        for (DA_UNSIGNED i = 0; i < byteData.size(); i++)
            delete [] byteData.at(i);
    #endif

    byteData.clear();

    #ifdef DA_DEBUG_MODE
        if (arraysAdded != 0)
        {
            DEngine::addFatal("DBytes::~DBytes(): leaked ");
            DEngine::addFatal(DEngine::intToStr(arraysAdded));
            DEngine::fatalError(" SafeArrays.");
        }//if arraysAdded
    #endif
}//destructor

DBytes& DBytes::operator= (const DBytes &source)
{
    if (this != &source)
    {
        clear();
        append(source);
    }//if this

    return *this;
}//=

void DBytes::setByte(DA_UNSIGNED pos,
                     uint8_t newByte)
{
    DA_UNSIGNED dataIndex = 0;//the index in byteData to access
    uint16_t arrayIndex = 0;//the index in the array to access

    calcIndexes(pos, dataIndex, arrayIndex);

    byteData.at(dataIndex)[arrayIndex] = newByte;
}//setByte

uint8_t DBytes::getByte(DA_UNSIGNED pos) const
{
    DA_UNSIGNED dataIndex = 0;//the index in byteData to access
    uint16_t arrayIndex = 0;//the index in the array to access

    calcIndexes(pos, dataIndex, arrayIndex);

    return byteData.at(dataIndex)[arrayIndex];
}//getByte

void DBytes::calcIndexes(DA_UNSIGNED pos,
                         DA_UNSIGNED &dataIndex,
                         uint16_t &arrayIndex) const
{
    #ifdef DA_DEBUG_MODE
        if (ARRAY_SIZE - startFilledBytes < 0)
        {
            DEngine::addFatal("DBytes::calcIndexes(): invalid startFilledBytes ");
            DEngine::fatalError(DEngine::intToStr(startFilledBytes));
        }//if ARRAY_SIZE - startFilledBytes
    #endif

    pos += (ARRAY_SIZE - startFilledBytes);

    if (pos > 0)
    {
        dataIndex = pos / ARRAY_SIZE;
        arrayIndex = pos % ARRAY_SIZE;
    }//if pos

    if (dataIndex >= byteData.size() - 1)
    {
        if (byteData.size() == 1)
        {
            if (static_cast<int>(arrayIndex) < ARRAY_SIZE - startFilledBytes)
            {
                DEngine::addFatal("DBytes::calcIndexes(): Invalid array index (start): ");
                DEngine::fatalError(DEngine::intToStr(arrayIndex));
            }//if arrayIndex
        }//if size()
        else
        {
            if (arrayIndex >= endFilledBytes)
            {
                DEngine::addFatal("DBytes::calcIndexes(): Invalid array index (end): ");
                DEngine::fatalError(DEngine::intToStr(arrayIndex));
            }//if arrayIndex
        }//else size()
    }//if dataIndex
}//calcIndexes

void DBytes::clear(void)
{
    #ifndef DA_DEBUG_MODE
        for (DA_UNSIGNED i = 0; i < byteData.size(); i++)
            delete byteData.at(i);
    #else
        arraysAdded -= byteData.size();
    #endif

    byteData.clear();

    startFilledBytes = 0;
    addElement();
}//clear

DA_UNSIGNED DBytes::size(void) const
{
    DA_UNSIGNED result = 0;//the size to return

    if (byteData.size() >= 2)
        result = (byteData.size() - 2) * ARRAY_SIZE;

    result += startFilledBytes + endFilledBytes;

    return result;
}//size

void DBytes::appendByte(uint8_t source)
{
    checkMaxSize(1);

    if ((endFilledBytes >= ARRAY_SIZE) || (byteData.size() == 1))
        addElement();

    byteData.at(byteData.size() - 1)[endFilledBytes] = source;
    endFilledBytes++;
}//appendByte

void DBytes::insertByte(uint8_t source)
{
    if (startFilledBytes >= ARRAY_SIZE)
    {
        if (byteData.size() == 1)
            endFilledBytes = ARRAY_SIZE;

        #ifdef DA_DEBUG_MODE
            byteData.push_front(DSafeArray(ARRAY_SIZE, true));
            byteData.at(0).makePermanent();
            arraysAdded++;
        #else
            byteData.push_front(new uint8_t[ARRAY_SIZE]);
        #endif

        startFilledBytes = 0;
    }//if startFilledBytes

    byteData.at(0)[ARRAY_SIZE - (startFilledBytes + 1)] = source;

    startFilledBytes++;
}//insertByte

void DBytes::removeFirst(DA_UNSIGNED numRemovals)
{
    //### Handle 'easy' cases ###
    if (numRemovals >= size())
        clear();
    else
    {
        if (numRemovals < startFilledBytes)
            startFilledBytes -= numRemovals;
        else
            //Handle the case where whole elements need to be removed
            performRemoveFirst(numRemovals);
    }//else numRemovals
}//removeFirst

void DBytes::performRemoveFirst(DA_UNSIGNED numRemovals)
{
    DA_UNSIGNED completeElements = 0;//the number of full elements in byteData to delete
    const DA_UNSIGNED origSize = size();//the original number of bytes in byteData
    const DA_UNSIGNED origNumRemovals = numRemovals;

    //### Remove the first element ###
    numRemovals -= startFilledBytes;
    if (origNumRemovals >= startFilledBytes)
        startFilledBytes = 0;
    else
        startFilledBytes -= origNumRemovals;

    //### Remove complete elements ###
    if (numRemovals > 0)
        completeElements = numRemovals / ARRAY_SIZE;

    for (DA_UNSIGNED i = 0; i < completeElements; i++)
    {
        numRemovals -= ARRAY_SIZE;

        #ifndef DA_DEBUG_MODE
            delete byteData.at(0);
        #else
            arraysAdded--;
        #endif
        byteData.pop_front();
    }//for i

    //### Prune the final element ###
    if (numRemovals > 0)
    {
        //Should be less than one element's worth of bytes to remove
        if (numRemovals >= ARRAY_SIZE)
            DEngine::fatalError("DBytes::removeFirst(): numRemovals > ARRAY_SIZE");

        //We already pruned the first element, something's wrong
        if (byteData.size() <= 1)
            DEngine::fatalError("DBytes::removeFirst(): byteData.size() <= 1");

        //Remove the first element, which should have zero bytes
        else
        {
            if (startFilledBytes != 0)
                DEngine::fatalError("DBytes::removeFirst(): startFilledBytes == 0");

            #ifndef DA_DEBUG_MODE
                delete byteData.at(0);
            #else
                arraysAdded--;
            #endif
            byteData.pop_front();
        }//else size()

        //Configure a new first element
        if (byteData.size() == 1)
        {
            if (numRemovals > endFilledBytes)
                DEngine::fatalError("DBytes::removeFirst(): numRemovals > endFilledBytes");

            convertLastToFirst(numRemovals);
        }//if size()
        else
            startFilledBytes = ARRAY_SIZE - numRemovals;
    }//if numRemovals

    if (size() != origSize - origNumRemovals)
    {
        DEngine::addFatal("DBytes::removeFirst(): removed wrong number of bytes! size: ");
        DEngine::addFatal(DEngine::intToStr(size()));
        DEngine::addFatal(" expected size: ");
        DEngine::fatalError(DEngine::intToStr(origSize - origNumRemovals));
    }//if size()
}//performRemoveFirst

void DBytes::convertLastToFirst(DA_UNSIGNED numRemovals)
{
    #ifdef DA_DEBUG_MODE
        DSafeArray oneElement(0, false);//the only element in dataBytes
    #else
        uint8_t *oneElement = NULL;//the only element in dataBytes
    #endif
    DA_UNSIGNED sourceStart = 0;//the indexes of the first and last array items to copy
    DA_UNSIGNED sourceEnd = 0;
    DA_UNSIGNED writeIndex = 0;//the index in oneElement to write to

    //### Calculate the number of bytes in the first element ###
    startFilledBytes = endFilledBytes - numRemovals;
    if (startFilledBytes > ARRAY_SIZE - numRemovals)
        DEngine::fatalError("DBytes::convertLastToFirst(): startFilledBytes too large");

    //### Zero endFilledBytes ###
    endFilledBytes = 0;

    //### Extract the first element ###
    oneElement = byteData.at(0);

    //### Calculate the range of bytes to copy ###
    sourceStart = startFilledBytes + numRemovals - 1;
    sourceEnd = numRemovals;

    //### Start writing at the bottom of the array ###
    writeIndex = ARRAY_SIZE - 1;

    //### Shuffle the bytes ###
    for (DA_UNSIGNED i = sourceStart; i >= sourceEnd; i--)
    {
        oneElement[writeIndex] = oneElement[i];
        writeIndex--;
    }//for i
}//convertLastToFirst

void DBytes::append(const DBytes &source)
{
    append(source, 0, source.size());
}//append

void DBytes::append(const DBytes &source,
                    DA_UNSIGNED index,
                    DA_UNSIGNED length)
{
    #ifdef DA_DEBUG_MODE
        DSafeArray readBuffer(0, false);//the element from the source to read from
        DSafeArray writeBuffer(0, false);//the element in byteData to write to
    #else
        uint8_t *readBuffer = NULL;//the element from the source to read from
        uint8_t *writeBuffer = NULL;//the element in byteData to write to
    #endif
    DA_UNSIGNED sourceElement = 0;//the first element in source to copy
    DA_UNSIGNED writeAmount = 0;//the number of bytes to copy
    DA_UNSIGNED destCapacity = 0;//the number of free bytes in writeElement
    uint16_t sourceArray = 0;//the first array index to copy

    checkMaxSize(length);

    if ((source.size() > 0) && (length > 0))
    {
        //### Find the index of the first element to copy ###
        source.calcIndexes(index, sourceElement, sourceArray);

        //### Create a new element ###
        if (byteData.size() == 1)
            addElement();

        while ((sourceElement < source.byteData.size()) && (length > 0))
        {
            //### Verify that the parameters are valid ###
            if (index < 0)
            {
                DEngine::addFatal("DBytes::append(): index is ");
                DEngine::fatalError(DEngine::intToStr(index));
            }//if index

            if (index + length > source.size())
            {
                DEngine::addFatal("DBytes::append(): index + length is ");
                DEngine::addFatal(DEngine::intToStr(index + length));
                DEngine::addFatal(", size is ");
                DEngine::fatalError(DEngine::intToStr(size()));
            }//if index + length

            //### Find the index of the first element to copy ###
            source.calcIndexes(index, sourceElement, sourceArray);

            //### Calculate the number of bytes we can copy from the source ###
            if (sourceElement == source.byteData.size() - 1)
                writeAmount = source.endFilledBytes - sourceArray;
            else
                writeAmount = ARRAY_SIZE - sourceArray;

            //### Calculate the number of bytes we can write to the dest ###
            destCapacity = ARRAY_SIZE - endFilledBytes;
            if (destCapacity == 0)
            {
                addElement();
                destCapacity = ARRAY_SIZE;
            }//if destCapacity

            //### Determine the final number of bytes to copy ###
            if (writeAmount > destCapacity)
                writeAmount = destCapacity;
            if (writeAmount > length)
                writeAmount = length;

            //### Extract the arrays to read and write to ###
            readBuffer = source.byteData.at(sourceElement);
            writeBuffer = byteData.at(byteData.size() - 1);

            //### Perform the copy ###
            #ifdef DA_DEBUG_MODE
                for (DA_UNSIGNED i = endFilledBytes; i < endFilledBytes + writeAmount; i++)
                {
                    writeBuffer[i] = readBuffer[sourceArray];
                    sourceArray++;
                }//for i
            #else
                memcpy(&writeBuffer[endFilledBytes], &readBuffer[sourceArray], writeAmount);
            #endif

            endFilledBytes += writeAmount;
            index += writeAmount;
            length -= writeAmount;
        }//while sourceElement && length

        if (length != 0)
        {
            DEngine::addFatal("DBytes::append(): bytes remaining to append: ");
            DEngine::fatalError(DEngine::intToStr(length));
        }//if length
    }//if size() && length

    //We know that source.size() must be zero, check if length is non-zero
    else if (length > 0)
    {
        DEngine::addFatal("DBytes::append(): can't copy ");
        DEngine::addFatal(DEngine::intToStr(length));
        DEngine::addFatal(" bytes from source, which only has ");
        DEngine::fatalError(DEngine::intToStr(source.size()));
    }//else if length
}//append

void DBytes::appendArray(const uint8_t *source,
                         DA_UNSIGNED length)
{
    #ifdef DA_DEBUG_MODE
        DSafeArray writeBuffer(0, false);//the element in byteData to write to
    #else
        uint8_t *writeBuffer = NULL;//the element in byteData to write to
    #endif
    DA_UNSIGNED sourceIndex = 0;//the index of the first byte in source to copy
    DA_UNSIGNED writeAmount = 0;//the number of bytes we can write to the destination buffer
    DA_UNSIGNED destCapacity = 0;//the number of free bytes in writeElement

    checkMaxSize(length);

    if (length > 0)
    {
        //### Create a new element if needed ###
        if (byteData.size() == 1)
            addElement();

        while (sourceIndex < length)
        {
            //### Create a new element if necessary ###
            destCapacity = ARRAY_SIZE - endFilledBytes;
            if (destCapacity == 0)
            {
                addElement();
                destCapacity = ARRAY_SIZE;
            }//if destCapacity

            //### Calculate the number of bytes we can copy to the dest ###
            writeAmount = length - sourceIndex;
            if (writeAmount > destCapacity)
                writeAmount = destCapacity;

            //### Copy the data ###
            writeBuffer = byteData.at(byteData.size() - 1);

            #ifdef DA_DEBUG_MODE
                for (DA_UNSIGNED i = endFilledBytes; i < endFilledBytes + writeAmount; i++)
                {
                    writeBuffer[i] = source[sourceIndex];
                    sourceIndex++;
                }//for i
            #else
                memcpy(&writeBuffer[endFilledBytes], &source[sourceIndex], writeAmount);
                sourceIndex += writeAmount;
            #endif

            endFilledBytes += writeAmount;
        }//while sourceIndex

        if (sourceIndex != length)
        {
            DEngine::addFatal("DBytes::appendArray(): copied ");
            DEngine::addFatal(DEngine::intToStr(sourceIndex));
            DEngine::addFatal(" out of ");
            DEngine::fatalError(DEngine::intToStr(length));
        }//if sourceIndex
    }//if length
}//appendArray

void DBytes::toArray(uint8_t *theArray,
                     DA_UNSIGNED maxBytes)
{
    vector<uint8_t*> byteArrays;//each array of bytes to copy
    vector<uint16_t> arraySizes;//the number of bytes in the corresponding element in theArrays
    DA_UNSIGNED writePos = 0;//the index in theArray to write to
    #ifdef DA_DEBUG_MODE
        DSafeArray dataElement(0, false);//the current buffer to send
    #else
        uint8_t *dataElement = NULL;//the current buffer to send
    #endif

    if (theArray == NULL)
        DEngine::fatalError("DBytes::toArray(): the array is NULL");

    if (byteData.size() > 0)
    {
        getArrays(byteArrays, arraySizes, maxBytes);

        for (DA_UNSIGNED i = 0; i < byteArrays.size(); i++)
        {
            dataElement = byteArrays.at(i);

            #ifdef DA_DEBUG_MODE
                for (DA_UNSIGNED j = 0; j < arraySizes.at(j); j++)
                {
                    theArray[writePos] = dataElement[j];
                    writePos++;
                }//for j
            #else
                memcpy(&theArray[writePos], &dataElement[0], arraySizes.at(i));
                writePos += arraySizes.at(i);
            #endif
        }//for i
    }//if size()
}//toArray

vector<boost::asio::const_buffer> DBytes::getSendBuffers(DA_UNSIGNED maxBytes) const
{
    vector<uint8_t*> theArrays;//each array of bytes to send
    vector<uint16_t> arraySizes;//the number of bytes in the corresponding element in theArrays
    vector<boost::asio::const_buffer> result;//the buffers to return
    #ifdef DA_DEBUG_MODE
        DSafeArray dataElement(0, false);//the current buffer to send
    #else
        uint8_t *dataElement = NULL;//the current buffer to send
    #endif

    if (byteData.size() > 0)
    {
        getArrays(theArrays, arraySizes, maxBytes);

        for (DA_UNSIGNED i = 0; i < theArrays.size(); i++)
        {
            dataElement = theArrays.at(i);
            result.push_back(boost::asio::buffer(&dataElement[0], arraySizes.at(i)));
        }//for i
    }//if size()

    return result;
}//getSendBuffers

void DBytes::getArrays(vector<uint8_t*> &theArrays,
                       vector<uint16_t> &arraySizes,
                       DA_UNSIGNED maxBytes) const
{
    uint16_t numFromCurrent = 0;//the number of bytes to copy from the current element
    #ifdef DA_DEBUG_MODE
        DSafeArray dataElement(0, false);//the current buffer to send
    #else
        uint8_t *dataElement = NULL;//the current buffer to send
    #endif

    theArrays.clear();
    arraySizes.clear();

    if (byteData.size() > 0)
    {
        //### Append the first buffer ###
        if (startFilledBytes > 0)
        {
            numFromCurrent = startFilledBytes;
            if (maxBytes < startFilledBytes)
                numFromCurrent = maxBytes;

            dataElement = byteData.at(0);
            theArrays.push_back(&dataElement[ARRAY_SIZE - startFilledBytes]);
            arraySizes.push_back(numFromCurrent);

            maxBytes -= numFromCurrent;
        }//if startFilledBytes

        //### Append the filled buffers ###
        for (DA_UNSIGNED i = 1; (i < byteData.size() - 1) && (maxBytes > 0); i++)
        {
            dataElement = byteData.at(i);

            numFromCurrent = ARRAY_SIZE;
            if (maxBytes < ARRAY_SIZE)
                numFromCurrent = maxBytes;

            theArrays.push_back(&dataElement[0]);
            arraySizes.push_back(numFromCurrent);

            maxBytes -= numFromCurrent;
        }//for i

        //### Append the last buffer ###
        if (byteData.size() > 1)
        {
            if (endFilledBytes > 0)
            {
                numFromCurrent = endFilledBytes;
                if (maxBytes < endFilledBytes)
                    numFromCurrent = maxBytes;

                dataElement = byteData.at(byteData.size() - 1);

                theArrays.push_back(&dataElement[0]);
                arraySizes.push_back(numFromCurrent);

                maxBytes -= numFromCurrent;
            }//if endFilledBytes
        }//if size()
    }//if size()
}//getArrays

void DBytes::checkMaxSize(DA_UNSIGNED newBytes)
{
    if (size() + newBytes > DEngine::MAX_UNSIGNED_VALUE)
    {
        DEngine::addFatal("DBytes::checkMaxSize(): can't store more than ");
        DEngine::addFatal(DEngine::intToStr(DEngine::MAX_UNSIGNED_VALUE));
        DEngine::fatalError(" bytes.");
    }//if size() + newBytes
}//checkMaxSize

void DBytes::addElement(void)
{
    #ifdef DA_DEBUG_MODE
        byteData.push_back(DSafeArray(ARRAY_SIZE, true));
        byteData.at(byteData.size() - 1).makePermanent();
        arraysAdded++;
    #else
        byteData.push_back(new uint8_t[ARRAY_SIZE]);
    #endif

    endFilledBytes = 0;
}//addElement

long int DBytes::getNumInstances(void)
{
    #ifdef DA_DEBUG_MODE
        return numInstances;
    #else
        return 0;
    #endif
}//getNumInstances
