/*
 * Copyright (C) 2011-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DClient.hpp"
#include "DEngine.hpp"
#include "DBuffer.hpp"

using namespace std;
using namespace libdasyne;

DClient::DClient(DEngine *newNetEngine)
{
    internalNetEngine = newNetEngine;
    clientSocket = new boost::asio::ip::tcp::socket(*internalNetEngine->getIoService());
    bufferSize = DEFAULT_BUFFER_SIZE;

    setConnectedState(false);
    initVars();
}//constructor

DClient::DClient(DEngine *newNetEngine,
                 unsigned int newBufferSize)
{
    internalNetEngine = newNetEngine;
    clientSocket = new boost::asio::ip::tcp::socket(*internalNetEngine->getIoService());

    bufferSize = newBufferSize;
    if (bufferSize < 1024)
        bufferSize = 1024;

    setConnectedState(false);
    initVars();
}//constructor

DClient::DClient(DEngine *newNetEngine,
                 boost::asio::ip::tcp::socket *newSocket)
{
    internalNetEngine = newNetEngine;
    clientSocket = newSocket;
    bufferSize = DEFAULT_BUFFER_SIZE;

    connectedState = false;
    checkConnected();
    initVars();
    startReceive();
}//constructor

DClient::DClient(DEngine *newNetEngine,
                 boost::asio::ip::tcp::socket *newSocket,
                 unsigned int newBufferSize)
{
    internalNetEngine = newNetEngine;
    clientSocket = newSocket;

    bufferSize = newBufferSize;
    if (bufferSize < 1024)
        bufferSize = 1024;

    connectedState = false;
    checkConnected();
    initVars();
    startReceive();
}//constructor

DClient::~DClient(void)
{
    if (!closeSocket())
        cout << getErrMsg() << endl;

    delete sendQueue;
    sendQueue = NULL;

    delete clientSocket;
    clientSocket = NULL;

    delete hostResolver;
    hostResolver = NULL;

    delete [] receiveBuffer;
    receiveBuffer = NULL;
}//destructor

void DClient::checkConnected(void)
{
    boost::system::error_code errorCode;//error code returned by boost
    bool currentState = false;//true if we're currently connected

    if (clientSocket->is_open())
    {
        clientSocket->remote_endpoint(errorCode);
        if (!errorCode)
            currentState = true;
    }

    setConnectedState(currentState);
}//checkConnected

void DClient::initVars(void)
{
    hostResolver = new boost::asio::ip::tcp::resolver(*internalNetEngine->getIoService());
    receiveBuffer = new unsigned char[bufferSize];
    sendQueue = new DBuffer;
    endPointIter = boost::asio::ip::tcp::resolver::iterator();
    isSending = false;
    delayEnabled = true;

    clearAllHandlers();
}//initVars

void DClient::clearAllHandlers(void)
{
    haveSendHandler = false;
    haveErrorHandler = false;
    haveReceiveHandler = false;
    haveConnectHandler = false;
}

void DClient::startConnect(const string &hostName,
                           uint16_t newPortNumber,
                           libdasyne::IpVersion ipVersion,
                           boost::function<void (std::string, libdasyne::NetworkCode)> newHandler)
{
    boost::asio::ip::tcp::resolver::query *nameQuery = NULL;//query to resolve hostName to an IP address
    string portString = DEngine::intToStr(newPortNumber);

    connectHandler = newHandler;
    haveConnectHandler = true;

    clientSocket->close();
    setConnectedState(false);

    if (ipVersion == DA_IPV4)
        nameQuery = new boost::asio::ip::tcp::resolver::query(boost::asio::ip::tcp::v4(), hostName, portString,
                                                              boost::asio::ip::resolver_query_base::numeric_service);
    else if (ipVersion == DA_IPV6)
        nameQuery = new boost::asio::ip::tcp::resolver::query(boost::asio::ip::tcp::v6(), hostName, portString,
                                                              boost::asio::ip::resolver_query_base::numeric_service);

    hostResolver->async_resolve(*nameQuery, boost::bind(&DClient::resolveFinished, this,
                                                        boost::asio::placeholders::error,
                                                        boost::asio::placeholders::iterator));

    delete nameQuery;
    nameQuery = NULL;
}//startConnect

void DClient::resolveFinished(const boost::system::error_code &errorCode,
                              boost::asio::ip::tcp::resolver::iterator iterator)
{
    boost::asio::ip::tcp::endpoint endpoint;//the IP address and port number to connect to

    if (haveConnectHandler)
    {
        endPointIter = iterator;

        if (errorCode)
        {
            errMsg = "Couldn't resolve hostname: ";
            errMsg.append(errorCode.message());
            connectHandler(errMsg, DA_HOSTNAME_ERROR);
        }
        else
        {
            endpoint = *endPointIter;
            endPointIter++;

            clientSocket->async_connect(endpoint, boost::bind(&DClient::connectFinished, this,
                                                              boost::asio::placeholders::error));
        }
    }
}//resolveFinished

void DClient::connectFinished(const boost::system::error_code &errorCode)
{
    boost::asio::ip::tcp::endpoint endpoint;//the IP address and port number to connect to
    bool allOK = true;//false on networking error

    if (haveConnectHandler)
    {
        //### Try a different endpoint if we couldn't connect ###
        if (errorCode)
        {
            if (errorCode != boost::asio::error::operation_aborted)
            {
                //Verify that there's another endpoint
                if (endPointIter == boost::asio::ip::tcp::resolver::iterator())
                {
                    allOK = false;
                    errMsg = "Couldn't connect: ";
                    errMsg.append(errorCode.message());

                    connectHandler(errMsg, DA_CONNECT_ERROR);
                }

                //Close the socket
                if (allOK)
                {
                    if (!closeSocket())
                    {
                        allOK = false;
                        connectHandler(errMsg, DA_CLOSE_ERROR);
                    }
                }

                //Start another asynchronous connect
                if (allOK)
                {
                    endpoint = *endPointIter;
                    endPointIter++;

                    clientSocket->async_connect(endpoint, boost::bind(&DClient::connectFinished, this,
                                                                    boost::asio::placeholders::error));
                }
            }
        }

        //### We connected successfully ###
        else
        {
            setConnectedState(true);
            startReceive();

            connectHandler("Success", DA_SUCCESS);
        }
    }
}//connectFinished

void DClient::send(const DBuffer &theData)
{
    if (!isConnected())
        reportError("DClient::send(): called, but the socket is not connected.", false);
    else
    {
        sendQueue->append(theData);
        sendBytes();
    }
}//send

void DClient::sendBytes(void)
{
    vector<boost::asio::const_buffer> sendBuffers;//the buffers to send
    int maxBytes = bufferSize;//the maximum number of bytes to send

    if (!isConnected())
        sendQueue->clear();

    if (!isSending)
    {
        //### Find the number of bytes to send ###
        if (sendQueue->size() < bufferSize)
            maxBytes = sendQueue->size();

        //### Send the data ###
        if (maxBytes > 0)
        {
            sendBuffers = sendQueue->getSendBuffers(maxBytes);

            isSending = true;
            clientSocket->async_send(sendBuffers,
                                     boost::bind(&DClient::sendFinished, this,
                                                 boost::asio::placeholders::error,
                                                 boost::asio::placeholders::bytes_transferred));
        }//if maxBytes

        //### Report that all data was sent ###
        else if (haveSendHandler)
            sendHandler();
    }//if !isSending
}//sendBytes

unsigned long int DClient::unsentBytes(void)
{
    return sendQueue->size();
}//unsentBytes

void DClient::sendFinished(const boost::system::error_code &errorCode,
                           size_t bytesSent)
{
    //### Verify that we're sending something ###
    if (!isSending)
        internalNetEngine->fatalError("DClient::sendFinished(): called, but we're not sending!");

    if (bytesSent >= 1)
        sendQueue->removeFirst(bytesSent);
    else
    {
        if (clientSocket->is_open())
        {
            setConnectedState(false);
            errMsg = "Sent zero bytes";

            reportError(errMsg, false);//closes socket
        }//if is_open()
    }//else bytesSent

    isSending = false;

    if (errorCode)
    {
        if (errorCode != boost::asio::error::operation_aborted)
            recoverableError(errorCode);
    }//if errorCode
    else
        sendBytes();
}//sendFinished

bool DClient::closeSocket(void)
{
    boost::system::error_code errorCode;//error code returned by an ASIO operation
    boost::system::error_condition theError;//convert errorCode into a portable value
    bool result = true;//false if we couldn't close the socket

    setConnectedState(false);

    if (clientSocket->is_open())
    {
        //### Shutdown the socket ###
        clientSocket->shutdown(boost::asio::ip::tcp::socket::basic_stream_socket::shutdown_both, errorCode);
        if (errorCode)
        {
            theError = errorCode.default_error_condition();

            if (theError.value() != boost::system::errc::not_connected)
            {
                errMsg = "Couldn't shutdown socket: ";
                errMsg.append(theError.message());
                result = false;
            }//if value()
        }//if errorCode

        //### Close the socket ###
        if (result)
        {
            clientSocket->close(errorCode);
            if (errorCode)
            {
                theError = errorCode.default_error_condition();

                errMsg = "Couldn't close socket: ";
                errMsg.append(theError.message());
                result = false;
            }//if errorCode
        }//if result
    }//if is_open()

    return result;
}//closeSocket

void DClient::startReceive(void)
{
    clientSocket->async_receive(boost::asio::buffer(receiveBuffer, bufferSize),
                                boost::bind(&DClient::receiveFinished, this,
                                            boost::asio::placeholders::error,
                                            boost::asio::placeholders::bytes_transferred));
}//startReceive

void DClient::receiveFinished(const boost::system::error_code &errorCode,
                              size_t bytesReceived)
{
    DBuffer receivedData;//the data received

    if (errorCode)
    {
        if (errorCode != boost::asio::error::operation_aborted)
            recoverableError(errorCode);
    }//if errorCode
    else
    {
        receivedData.appendArray(receiveBuffer, bytesReceived);

        if (receivedData.size() > 0)
        {
            if (haveReceiveHandler)
                receiveHandler(receivedData);
        }//if size()

        else
        {
            if (clientSocket->is_open())
            {
                setConnectedState(false);
                errMsg = "Received zero bytes.";

                reportError(errMsg, false);//closes socket
            }//if is_open()
        }//else size()

        if (connectedState)
            startReceive();
    }//else
}//receiveFinished

string DClient::getLocalAddress(void)
{
    boost::system::error_code errorCode;//an error code from calling local_endpoint()
    boost::system::error_condition theError;//convert errorCode to a portable value
    boost::asio::ip::tcp::endpoint localEndpoint;//the local endpoint for this DClient
    boost::asio::ip::address localAddress;//the local address of this DClient
    string result;//the address to return

    localEndpoint = clientSocket->local_endpoint(errorCode);
    if (errorCode)
    {
        theError = errorCode.default_error_condition();
        errMsg = theError.message();
        errMsg.append(": ");
        errMsg.append(DEngine::intToStr(theError.value()));
    }//if errorCode
    else
    {
        localAddress = localEndpoint.address();
        result = localAddress.to_string();
    }//else errorCode

    return result;
}//getLocalAddress

std::string DClient::getRemoteAddress(void)
{
    return remoteAddress;
}//getRemoteAddress

void DClient::recoverableError(const boost::system::error_code &errorCode)
{
    #ifndef DA_WINDOWS_VERSION
        boost::system::error_condition theError = errorCode.default_error_condition();
    #endif
    string errMsg;//the error to report

    #ifdef DA_WINDOWS_VERSION
        errMsg = errorCode.message();
        errMsg.append(": ");
        errMsg.append(DEngine::intToStr(errorCode.value()));
	#else
        errMsg = theError.message();
        errMsg.append(": ");
        errMsg.append(DEngine::intToStr(theError.value()));
    #endif

    reportError(errMsg, false);
}//recoverableError

void DClient::reportError(const std::string &errMsg,
                          bool socketOK)
{
    //### Report recoverable errors ###
    if (socketOK)
    {
        if (haveErrorHandler)
            errorHandler(errMsg, socketOK);
    }//if socketOK

    //### Report unrecoverable errors ###
    else
    {
        setConnectedState(false);

        if (clientSocket->is_open())
        {
            if (haveErrorHandler)
            {
                errorHandler(errMsg, socketOK);
                if ((!closeSocket()) && (haveErrorHandler))
                    errorHandler(errMsg, socketOK);
            }
            else
                closeSocket();
        }//if is_open()
    }//else socketOK
}//reportError

bool DClient::setDelay(bool newValue)
{
    boost::asio::ip::tcp::no_delay option(!newValue);
    boost::system::error_code errorCode;//ASIO error code returned by set_option
    bool result = true;//false if we couldn't set the option

    clientSocket->set_option(option, errorCode);
    if (errorCode)
    {
        errMsg = "Couldn't set option: ";
        errMsg.append(errorCode.message());
        result = false;
    }//if errorCode
    else
        delayEnabled = newValue;

    delayToggled();

    return result;
}//setDelay

bool DClient::getDelayEnabled(void)
{
    return delayEnabled;
}//getDelayEnabled

bool DClient::isConnected(void)
{
    return connectedState;
}//isConnected

void DClient::delayToggled(void)
{
}//delayToggled

unsigned int DClient::getBufferSize(void)
{
    return bufferSize;
}//getBufferSize

void DClient::setSendHandler(boost::function<void (void)> newHandler)
{
    sendHandler = newHandler;
    haveSendHandler = true;
}//setSendHandler

void DClient::setReceiveHandler(boost::function<void (libdasyne::DBuffer&)> newHandler)
{
    receiveHandler = newHandler;
    haveReceiveHandler = true;
}//setReceiveHandler

void DClient::setErrorHandler(boost::function<void (std::string, bool)> newHandler)
{
    errorHandler = newHandler;
    haveErrorHandler = true;
}//setErrorHandler

void DClient::extractAddress(void)
{
    boost::system::error_code errorCode;//an error code from calling local_endpoint()
    boost::system::error_condition theError;//convert errorCode to a portable value
    boost::asio::ip::tcp::endpoint remoteEndpoint;//the remote host's endpoint
    boost::asio::ip::address boostAddress;//the remote host's IP address

    if (clientSocket->is_open())
    {
        remoteEndpoint = clientSocket->remote_endpoint(errorCode);
        if (errorCode)
        {
            theError = errorCode.default_error_condition();
            errMsg = theError.message();
            errMsg.append(": ");
            errMsg.append(DEngine::intToStr(theError.value()));
        }//if errorCode
        else
        {
            boostAddress = remoteEndpoint.address();
            remoteAddress = boostAddress.to_string();
        }//else errorCode
    }//if is_open()
}//extractAddress

void DClient::setConnectedState(bool newValue)
{
    connectedState = newValue;

    if (connectedState)
        extractAddress();
    else
        remoteAddress.clear();
}//setConnectedState

string DClient::getErrMsg(void)
{
    return errMsg;
}//getErrMsg
