/*
 * Copyright (C) 2011-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    A TCP connection, created either by connecting to a server, or accepting a connection
    as a server. To make use of the DClient's events, the programmer should make DClient child
    class. The connectHandler, receiveHandler, sendHandler and errorHandler functions
    will be called when the appropriate event occurs, and can be implemented in the child to
    respond to them.
*/

#ifndef DCLIENT_HPP_INCLUDED
#define DCLIENT_HPP_INCLUDED

#include "DEngine.hpp"
#include "DSocket.hpp"

namespace libdasyne
{
    class DBuffer;

    class DClient : public DSocket
    {
        public:
            //constructors
            DClient(DEngine *newNetEngine);
            DClient(DEngine *newNetEngine,
                    unsigned int newBufferSize);

            //constructor, should be called when a DServer accepts a new connection
            DClient(DEngine *newNetEngine,
                    boost::asio::ip::tcp::socket *newSocket);
            DClient(DEngine *newNetEngine,
                    boost::asio::ip::tcp::socket *newSocket,
                    unsigned int newBufferSize);

            //destructor
            virtual ~DClient(void);

            //Start a connection to the specified host name and port.
            //The event handler will called when the connect operation completes. status will have one of the
            //following values: DA_SUCCESS, DA_HOSTNAME_ERROR, DA_CLOSE_ERROR, DA_CONNECT_ERROR
            void startConnect(const std::string &hostName,
                              uint16_t newPortNumber,
                              IpVersion ipVersion,
                              boost::function<void (std::string, libdasyne::NetworkCode)> newHandler);

            //Sends the specified data to the remote host
            virtual void send(const libdasyne::DBuffer &theData);

            //Closes the socket, automatically called by the destructor. Returns false on errors and
            //sets errMsg.
            bool closeSocket(void);

            //returns a description of the most recent error
            std::string getErrMsg(void);

            //Returns this  client's local address, or an empty string if we couldn't obtain it
            std::string getLocalAddress(void);

            //Returns the address of the remote host, or an empty string if it couldn't be obtained.
            //Say, if this socket isn't connected.
            std::string getRemoteAddress(void);

            //Returns the size of the send and receive buffers in bytes
            unsigned int getBufferSize(void);

            //Enable or disable Nagle's algorithm. By default, Nagle's algorithm is enabled and
            //data is not immediately sent. Small messages are buffered, then sent together, improving
            //network efficiency but increasing latency. Returns false if the delay couldn't be changed
            //and sets errMsg.
            bool setDelay(bool newValue);

            //Returns true if this socket is connected, false otherwise
            bool isConnected(void);

            //Returns the number of unsent bytes in the DClient's send buffer
            virtual unsigned long int unsentBytes(void);

            //Event handler, called when there are no more bytes to send
            void setSendHandler(boost::function<void (void)> newHandler);

            //Event handler, called when we've received data
            void setReceiveHandler(boost::function<void (DBuffer&)> newHandler);

            //Event handler, called when an error occurs. If socketOK is false, the connection was lost
            void setErrorHandler(boost::function<void (std::string, bool)> newHandler);

            //Removes all user-defined handlers
            virtual void clearAllHandlers(void);

            //The default number of bytes we'll send or receive at one time
            static const unsigned int DEFAULT_BUFFER_SIZE = 16384;

        protected:
            //Handles events for this object
            DEngine *internalNetEngine;

            //a description of the most recent error
            std::string errMsg;

            //Returns true if Nagle's algorithm is enabled
            bool getDelayEnabled(void);

            //Report that setDelay was called
            virtual void delayToggled(void);

            //Calls errorHandler and closes the socket if socketOK is false
            void reportError(const std::string &errMsg,
                             bool socketOK);

        private:
            //Data that's waiting to be sent to the remote host
            DBuffer *sendQueue;

            //resolves host names into a boost endpoint
            boost::asio::ip::tcp::resolver *hostResolver;

            //Our connection to the server
            boost::asio::ip::tcp::socket *clientSocket;

            //An array of bytes to hold received data
            unsigned char *receiveBuffer;

            //The address of the remote host, or an empty string if not connected
            std::string remoteAddress;

            //The hostname may resolve to multiple end points, which can be iterated through with
            //endPointIter.
            boost::asio::ip::tcp::resolver::iterator endPointIter;

            //Pointers to event handlers
            boost::function<void (std::string, NetworkCode)> connectHandler;
            boost::function<void (void)> sendHandler;
            boost::function<void (std::string, bool)> errorHandler;
            boost::function<void (DBuffer&)> receiveHandler;

            //The maximum number of bytes to send or receive at one time
            unsigned int bufferSize;

            //True if we're currently sending data, false if idle
            bool isSending;

            //True if Nagle's algorithm is enabled
            bool delayEnabled;

            //True if we're currently connected, false otherwise. Should be modified through
            //setConnectedState().
            bool connectedState;

            //True if we have the corresponding event handler
            bool haveSendHandler;
            bool haveErrorHandler;
            bool haveReceiveHandler;
            bool haveConnectHandler;

            //############### Functions ###############

            //Called when the hostname is resolved to an IP address, or if it couldn't
            //be resolved
            void resolveFinished(const boost::system::error_code &errorCode,
                                 boost::asio::ip::tcp::resolver::iterator iterator);

            //Called when a connection attempt completes
            void connectFinished(const boost::system::error_code &errorCode);

            //Called when a send operation completes
            void sendFinished(const boost::system::error_code &errorCode,
                              std::size_t bytesSent);

            //Called when a receive operation completes
            void receiveFinished(const boost::system::error_code &errorCode,
                                 std::size_t bytesReceived);

            //Extracts and sends up to SEND_AMOUNT bytes from the sendQueue
            void sendBytes(void);

            //Initialize member variables to values used by both constructors
            void initVars(void);

            //Start an asynchronous receive operation
            void startReceive(void);

            //Calls errorHandler with a description of the problem, with socketOK set to true
            void recoverableError(const boost::system::error_code &errorCode);

            //Attempts to determine if clientSocket is connected. Boost doesn't provide a direct
            //way to tell, but we can check if the socket is open and if there's a remote endpoint.
            //Sets the connected variable appropriately.
            void checkConnected(void);

            //Set the value of connectedState, and either clears remoteAddress or assigns it to the proper value
            void setConnectedState(bool newValue);

            //Sets remoteAddress to the remote host's IP address
            void extractAddress(void);

    };//DClient
}//namespace libdasyne

#endif
