/*
 * Copyright (C) 2011-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DDispatcher.hpp"
#include "DEventMap.hpp"
#include "DBuffer.hpp"
#include "DBufferIter.hpp"
#include "DFlushEvent.hpp"
#include "DUdpModule.hpp"

using namespace std;
using namespace libdasyne;
using namespace libdasyne::DasyneInt;

DDispatcher::DDispatcher(DEngine *newNetEngine) : DClient(newNetEngine)
{
    initVars();
}//constructor

DDispatcher::DDispatcher(DEngine *newNetEngine,
                         unsigned int newBufferSize) : DClient(newNetEngine, newBufferSize)
{
    initVars();
}//cosntructor

DDispatcher::DDispatcher(DEngine *newNetEngine,
                         boost::asio::ip::tcp::socket *newSocket) : DClient(newNetEngine, newSocket)
{
    initVars();
}//constructor

DDispatcher::DDispatcher(DEngine *newNetEngine,
                         boost::asio::ip::tcp::socket *newSocket,
                         unsigned int newBufferSize) : DClient(newNetEngine, newSocket, newBufferSize)
{
    initVars();
}//constructor

DDispatcher::~DDispatcher(void)
{
    for (unsigned int i = 0; i < subtypeMaps.size(); i++)
        delete subtypeMaps.at(i);
    subtypeMaps.clear();

    clearOutMessages();
    disableUdp();
}//destructor

void DDispatcher::initVars(void)
{
    createdCustomHandlers = false;
    myState = DADD_FIRST_HEADER;
    recvSize = 0;
    eventHandler = NULL;
    streamOK = true;
    junkAmount = 0;
    haveSingleHandler = false;
    haveUdpRequestHandler = false;
    sizeFieldLength = 0;
    numMsgHandlers = 0;
    udpModule = NULL;
    remoteUdpPort = 0;
    waitingForUdpReply = false;
    udpSendKey = 0;
    udpOK = false;

    for (unsigned int i = 0; i < 256; i++)
        subtypeMaps.push_back(new DEventMap(i));

    setReceiveHandler(boost::bind(&DDispatcher::receiveTcpData, this, _1));
}//initVars

void DDispatcher::dispatch(DBuffer *newMessage)
{
    //### Send the message immediately ###
    if (getDelayEnabled())
    {
        DClient::send(*newMessage);

        delete newMessage;
        newMessage = NULL;
    }//if getDelayEnabled()

    //### Wait for a flush before sending ###
    else
    {
        outMessages.push_back(newMessage);
        newMessage = NULL;
    }//else getDelayEnabled()
}//dispatch

void DDispatcher::sendTcp(const DBuffer &theData)
{
    sendTcp(DEFAULT_MESSAGE_TYPE, theData);
}//sendTcp

void DDispatcher::sendTcp(uint8_t messageType,
                          const DBuffer &theData)
{
    DBuffer *newMessage = new DBuffer;//the new message to send

    //### Verify that messageType isn't reserved ###
    checkReservedType(messageType, true);

    //### Construct the message ###
    appendTcpHeader(newMessage);
    appendMessageSize(theData, newMessage);

    newMessage->appendByte(messageType);
    newMessage->append(theData);

    dispatch(newMessage);
}//sendTcp

void DDispatcher::sendTcp(uint8_t messageType,
                          uint16_t subType,
                          const DBuffer &theData)
{
    DBuffer *newMessage = new DBuffer;//the new message to send

    //### Verify that messageType isn't reserved ###
    checkReservedType(messageType, true);

    //### Construct the message ###
    appendTcpHeader(newMessage);
    appendMessageSize(theData, newMessage);

    newMessage->appendByte(messageType);
    newMessage->appendVariable(subType);
    newMessage->append(theData);

    dispatch(newMessage);
}//sendTcp

void DDispatcher::sendCustomTcp(uint16_t subType,
                                const DBuffer &theData)
{
    DBuffer *newMessage = new DBuffer;//the new message to send

    appendTcpHeader(newMessage);
    appendMessageSize(theData, newMessage);

    newMessage->appendByte(RESERVED_MESSAGE_TYPE);
    newMessage->appendVariable(subType);
    newMessage->append(theData);

    dispatch(newMessage);
}//sendCustomTcp

void DDispatcher::appendMessageSize(const DBuffer &theData,
                                    DBuffer *newMessage)
{
    uint32_t fourByteSize = 0;//the maximum message size that fits in four bytes

    //### Calculate the max message size that fits in four bytes ###
    fourByteSize = 2147483647;
    fourByteSize *= 2;
    fourByteSize++;

    //### Add the message size ###
    if (theData.size() == 0)
        newMessage->appendByte(0);

    else if (theData.size() <= 255)
    {
        newMessage->appendByte(1);
        newMessage->appendVariable(static_cast<uint8_t>(theData.size()));
    }//else if size()

    else if (theData.size() <= 65535)
    {
        newMessage->appendByte(2);
        newMessage->appendVariable(static_cast<uint16_t>(theData.size()));
    }//else if size()

    else if (theData.size() <= fourByteSize)
    {
        newMessage->appendByte(4);
        newMessage->appendVariable(static_cast<uint32_t>(theData.size()));
    }//else if size()

    else
    {
        newMessage->appendByte(8);
        newMessage->appendVariable(static_cast<uint64_t>(theData.size()));
    }//else size()
}//appendMessageSize

void DDispatcher::appendTcpHeader(DBuffer *newMessage)
{
    newMessage->appendByte(FIRST_HEADER_BYTE);
    newMessage->appendByte(SECOND_HEADER_BYTE);
    newMessage->appendByte(THIRD_HEADER_BYTE);
}//appendHeader

void DDispatcher::addHandler(boost::function<void (DBuffer&)> newHandler)
{
    if (numMsgHandlers > 0)
    {
        DEngine::addFatal("Tried to add a generic event handler, but we already ");
        DEngine::fatalError("have handlers for specific message types.");
    }//if numMsgHandlers

    singleHandler = newHandler;
    haveSingleHandler = true;
}//addHandler

void DDispatcher::addHandler(uint8_t messageType,
                             boost::function<void (DBuffer&)> newHandler)
{
    if (haveSingleHandler)
    {
        DEngine::addFatal("Tried to add a handler for message type ");
        DEngine::addFatal(DEngine::intToStr(messageType));
        DEngine::fatalError(", but we already have a handler for all types of messages.");
    }//if haveSingleHandler

    //### Verify that messageType isn't reserved ###
    checkReservedType(messageType, false);

    subtypeMaps.at(messageType)->addHandler(newHandler);
    numMsgHandlers++;
}//addHandler

void DDispatcher::addHandler(uint8_t messageType,
                             uint16_t subType,
                             boost::function<void (DBuffer&)> newHandler)
{
    if (haveSingleHandler)
    {
        DEngine::addFatal("Tried to add a handler for message type ");
        DEngine::addFatal(DEngine::intToStr(messageType));
        DEngine::fatalError(", but we already have a handler for all types of messages.");
    }//if haveSingleHandler

    //### Verify that messageType isn't reserved ###
    checkReservedType(messageType, false);

    subtypeMaps.at(messageType)->addHandler(subType, newHandler);
    numMsgHandlers++;
}//addHandler

void DDispatcher::checkReservedType(uint8_t messageType,
                                    bool isSending)
{
    if (messageType == RESERVED_MESSAGE_TYPE)
    {
        if (isSending)
            DEngine::addFatal("Tried to send a message with type ");
        else
            DEngine::addFatal("Tried to add a handler for message type ");

        DEngine::addFatal(DEngine::intToStr(RESERVED_MESSAGE_TYPE));
        DEngine::fatalError(", but that type is reserved for internal use.");
    }//if messageType
}//checkReservedType

void DDispatcher::removeHandler(void)
{
    if (!haveSingleHandler)
        DEngine::fatalError("Tried to remove the handler for all received messages, but there isn't one");

    haveSingleHandler = false;

    if (numMsgHandlers > 0)
        DEngine::fatalError("Removed handler for all received messages, but there are also subtypes!?");
}//removeHandler

void DDispatcher::removeHandler(uint8_t messageType)
{
    verifySubtypesExist(messageType);

    subtypeMaps.at(messageType)->removeHandler();
    numMsgHandlers--;
}//removeHandler

void DDispatcher::removeHandler(uint8_t messageType,
                                uint16_t subType)
{
    verifySubtypesExist(messageType);

    if (subtypeMaps.at(messageType)->removeHandler(subType))
        numMsgHandlers--;
}//removeHandler

void DDispatcher::verifySubtypesExist(uint8_t messageType)
{
    //### Verify that we aren't using a single handler ###
    if (haveSingleHandler)
    {
        DEngine::addFatal("Tried to remove handler for message type ");
        DEngine::addFatal(DEngine::intToStr(messageType));
        DEngine::fatalError(", but we have a single handler for all types.");
    }//if haveSingleHandler

    //### Verify that at least one message-specific handler was defined ###
    if (numMsgHandlers == 0)
    {
        DEngine::addFatal("Tried to remove handler for message type ");
        DEngine::addFatal(DEngine::intToStr(messageType));
        DEngine::fatalError(", but no handlers have been defined for specific message types.");
    }//if numMsgHandlers
}//verifySubtypesExist

void DDispatcher::receiveTcpData(DBuffer &receivedBytes)
{
    DBufferIter recvIter;//iterates through the received bytes
    unsigned int byteIndex = 0;//the index of the current byte to extract from receivedBytes
    int copyAmount = 0;//the number of bytes to copy from receivedBytes
    int bytesAvailable = 0;//the number of bytes remaining in receivedBytes
    uint16_t temp = 0;//stores a two-byte parameter from receivedBytes
    uint8_t oneByte = 0;//a single byte from receivedBytes

    if (!streamOK)
        junkAmount += receivedBytes.size();

    recvIter.setBuffer(&receivedBytes);

    while (!recvIter.finished())
    {
        oneByte = recvIter.value();

        switch (myState)
        {
            case DADD_FIRST_HEADER:
                if (checkHeaderByte(FIRST_HEADER_BYTE, oneByte))
                    myState = DADD_SECOND_HEADER;
            break;

            case DADD_SECOND_HEADER:
                if (checkHeaderByte(SECOND_HEADER_BYTE, oneByte))
                    myState = DADD_THIRD_HEADER;
            break;

            case DADD_THIRD_HEADER:
                if (checkHeaderByte(THIRD_HEADER_BYTE, oneByte))
                {
                    myState = DADD_SIZE_LENGTH;
                    streamOK = true;
                    junkAmount = 0;
                }//if checkHeaderByte
            break;

            case DADD_SIZE_LENGTH:
                sizeFieldLength = oneByte;

                if (sizeFieldLength == 0)
                {
                    recvSize = 0;
                    recvMessage.clear();
                    myState = DADD_MESSAGE_TYPE;
                }//if sizeFieldLength

                else if ((sizeFieldLength != 1)
                      && (sizeFieldLength != 2)
                      && (sizeFieldLength != 4)
                      && (sizeFieldLength != 8))
                {
                    errMsg = "Invalid size for the length value: ";
                    errMsg.append(DEngine::intToStr(sizeFieldLength));
                    reportError(errMsg, true);
                    myState = DADD_FIRST_HEADER;
                }//if sizeFieldLength ... && sizeFieldLength

                else
                {
                    dataField.clear();
                    myState = DADD_SIZE;
                }//else sizeFieldLength ... && sizeFieldLength
            break;

            case DADD_SIZE:
                extractMessageLength(oneByte);
            break;

            case DADD_MESSAGE_TYPE:
                extractMessageType(oneByte);
            break;

            case DADD_FIRST_SUBTYPE:
                dataField.appendByte(oneByte);
                myState = DADD_SECOND_SUBTYPE;
            break;

            case DADD_SECOND_SUBTYPE:
                dataField.appendByte(oneByte);
                dataField.extractVariable(0, temp);

                if (subtypeMaps.at(recvMsgType)->getHandler(temp, currentHandler))
                {
                    if (recvSize == 0)
                    {
                        currentHandler(recvMessage);
                        myState = DADD_FIRST_HEADER;
                    }//if recvSize
                    else
                        myState = DADD_MESSAGE_BODY;
                }//if getHandler()
                else
                {
                    errMsg = "No handler set for message type ";
                    errMsg.append(DEngine::intToStr(recvMsgType));
                    errMsg.append(" sub-type ");
                    errMsg.append(DEngine::intToStr(temp));
                    reportError(errMsg, true);

                    myState = DADD_FIRST_HEADER;
                    streamOK = false;
                }//else getHandler
            break;

            case DADD_MESSAGE_BODY:
                //### Calculate the number of bytes to copy ###
                copyAmount = recvSize - recvMessage.size();
                if (copyAmount <= 0)
                    DEngine::fatalError("DDispatcher::receiveData(): received too many bytes!");

                //### Reduce it to the number of bytes available ###
                bytesAvailable = receivedBytes.size() - byteIndex;
                if (bytesAvailable < copyAmount)
                    copyAmount = bytesAvailable;

                //### Copy the bytes ###
                recvMessage.append(receivedBytes, byteIndex, copyAmount);
                byteIndex += copyAmount - 1;
                recvIter.setIndex(byteIndex);

                //### Report that we have a complete message ###
                if (recvMessage.size() == recvSize)
                {
                    myState = DADD_FIRST_HEADER;
                    currentHandler(recvMessage);
                }//if size()
                else if (recvMessage.size() > recvSize)
                    DEngine::fatalError("DDispatcher::receiveData(): stored too many bytes?!");
            break;

            default:
                DEngine::fatalError("DDispatcher::receiveData(): reached an unknown state");
                break;
        };//switch myState

        recvIter.next();
        byteIndex++;
    }//while !finished()

    if (!streamOK)
    {
        while (junkAmount > 65541)//64436 + 5 bytes header size
        {
            junkAmount -= 65541;
            errMsg = "Received 64 KB of data without finding a valid message.";
            reportError(errMsg, true);
        }//while junkAmount
    }//if !streamOK
}//receiveTcpData

void DDispatcher::extractMessageType(uint8_t oneByte)
{
    //### Create custom event handlers, if necessary ###
    if (oneByte == RESERVED_MESSAGE_TYPE)
    {
        addCustomHandlers();

        recvMsgType = oneByte;
        dataField.clear();
        myState = DADD_FIRST_SUBTYPE;
    }//if oneByte && !createdCustomHandlers

    //### Handle generic messages ###
    if ((haveSingleHandler) && (oneByte != RESERVED_MESSAGE_TYPE))
    {
        if (recvSize == 0)
        {
            singleHandler(recvMessage);
            myState = DADD_FIRST_HEADER;
        }//if recvSize
        else
        {
            currentHandler = singleHandler;
            myState = DADD_MESSAGE_BODY;
        }//else recvSize
    }//if haveSingleHandler && oneByte

    //### Handle message types, with no subtype ###
    else if (subtypeMaps.at(oneByte)->getHaveSingle())
    {
        currentHandler = subtypeMaps.at(oneByte)->getSingleHandler();

        if (recvSize == 0)
        {
            currentHandler(recvMessage);
            myState = DADD_FIRST_HEADER;
        }//if recvSize
        else
            myState = DADD_MESSAGE_BODY;
    }//if getHaveSingle()

    //### Handle subtypes ###
    else if (subtypeMaps.at(oneByte)->size() > 0)
    {
        recvMsgType = oneByte;
        dataField.clear();
        myState = DADD_FIRST_SUBTYPE;
    }//else if size()

    //### Report an error if no handler was found ###
    else
    {
        errMsg = "No handler set for message type ";
        errMsg.append(DEngine::intToStr(oneByte));
        reportError(errMsg, true);
        myState = DADD_FIRST_HEADER;
        streamOK = false;
    }//else size()
}//extractMessageType

void DDispatcher::extractMessageLength(uint8_t oneByte)
{
    uint8_t size8 = 0;//store the message length in a variable of the appropriate size
    uint16_t size16 = 0;
    uint32_t size32 = 0;
    uint64_t size64 = 0;

    dataField.appendByte(oneByte);
    if (dataField.size() >= sizeFieldLength)
    {
        if (sizeFieldLength == 1)
        {
            dataField.extractVariable(0, size8);
            recvSize = size8;
        }//if sizeFieldLength

        else if (sizeFieldLength == 2)
        {
            dataField.extractVariable(0, size16);
            recvSize = size16;
        }//else if sizeFieldLength

        else if (sizeFieldLength == 4)
        {
            dataField.extractVariable(0, size32);
            recvSize = size32;
        }//else if sizeFieldLength

        else if (sizeFieldLength == 8)
        {
            dataField.extractVariable(0, size64);
            recvSize = size64;
        }//else if sizeFieldLength

        else
            DEngine::fatalError("DDispatcher::extractMessageLength(): invalid sizeFieldLength");

        recvMessage.clear();
        currentHandler = NULL;
        myState = DADD_MESSAGE_TYPE;
    }//if size()
}//extractMessageLength

bool DDispatcher::checkHeaderByte(uint8_t expected,
                                  uint8_t received)
{
    bool result = true;//false if we received an invalid header byte

    if (received != expected)
    {
        myState = DADD_FIRST_HEADER;
        result = false;

        if (streamOK)
        {
            errMsg = "Expected header byte ";
            errMsg.append(DEngine::intToStr(expected));
            errMsg.append(", got ");
            errMsg.append(DEngine::intToStr(received));

            reportError(errMsg, true);
            streamOK = false;
        }//if streamOK
    }//if received

    return result;
}//checkHeaderByte

void DDispatcher::addCustomHandlers(void)
{
    DEventMap *reservedMap = subtypeMaps.at(RESERVED_MESSAGE_TYPE);

    if (!createdCustomHandlers)
    {
        reservedMap->addHandler(DADD_UDP_REQUEST, boost::bind(&DDispatcher::udpRequestEvent, this, _1));
        numMsgHandlers++;

        reservedMap->addHandler(DADD_UDP_RESULT, boost::bind(&DDispatcher::udpResultEvent, this, _1));
        numMsgHandlers++;

        createdCustomHandlers = true;
    }
}//addCustomHandlers

void DDispatcher::delayToggled(void)
{
    if (getDelayEnabled())
    {
        delete eventHandler;
        eventHandler = NULL;
    }//if getDelayEnabled
    else
    {
        if (eventHandler == NULL)
            eventHandler = new DFlushEvent(internalNetEngine, this);
    }//else getDelayEnabled
}//delayToggled

bool DDispatcher::flush(void)
{
    bool result = false;//true if there was data to flush

    if (outMessages.size() > 0)
        result = true;

    for (unsigned int i = 0; i < outMessages.size(); i++)
        DClient::send(*(outMessages.at(i)));
    clearOutMessages();

    return result;
}//flush

unsigned long int DDispatcher::unsentBytes(void)
{
    unsigned long int result = 0;//the number of bytes remaining to send

    result = DClient::unsentBytes();

    for (unsigned long int i = 0; i < outMessages.size(); i++)
        result += outMessages.at(i)->size();

    return result;
}//unsentBytes

void DDispatcher::clearOutMessages(void)
{
    for (unsigned long int i = 0; i < outMessages.size(); i++)
        delete outMessages.at(i);
    outMessages.clear();
}//clearOutMessages

void DDispatcher::clearAllHandlers(void)
{
    cout << "DDispatcher::clearAllHandlers(): called" << endl;
    disableUdp();
    DClient::clearAllHandlers();
}

//###################################
//############### UDP ###############
//###################################

bool DDispatcher::verifyUdpIsReady()
{
    bool result = true;//false if UDP communication is not enabled

    if (udpModule == NULL)
    {
        reportError("The DDispatcher does not have a DUdpModule", true);
        result = false;
    }

    if ((result) && (!udpOK))
    {
        reportError("The DDispatcher has a DUdpModule, but UDP is not yet enabled.", true);
        result = false;
    }

    return result;
}

void DDispatcher::sendUdp(const DBuffer &theData)
{
    if (verifyUdpIsReady())
        sendUdp(0, theData);
}//sendUdp

void DDispatcher::sendUdp(uint8_t messageType,
                          const DBuffer &theData)
{
    DBuffer newMessage;//the message to send

    if (verifyUdpIsReady())
    {
        newMessage.appendVariable(udpSendKey);
        newMessage.appendByte(messageType);
        newMessage.append(theData);

        udpModule->send(newMessage, getRemoteAddress(), remoteUdpPort);
    }
}//sendUdp

void DDispatcher::sendUdp(uint8_t messageType,
                          uint16_t subType,
                          const DBuffer &theData)
{
    DBuffer newMessage;//the message to send

    if (verifyUdpIsReady())
    {
        newMessage.appendVariable(udpSendKey);

        newMessage.appendByte(messageType);
        newMessage.appendVariable(subType);
        newMessage.append(theData);

        udpModule->send(newMessage, getRemoteAddress(), remoteUdpPort);
    }
}//sendUdp

void DDispatcher::receiveUdpData(DBuffer &receivedBytes)
{
    DEventMap* msgEventMap = NULL;//the sub-type map with the correct handler, if any
    boost::function<void (DBuffer&)> udpDataHandler;//the user-defined method to call
    uint16_t messageSubType = 0;//the subtype of the received message if any
    uint8_t messageType = 0;//the type of message we've received
    bool finished = false;//true if we've handled the message
    bool allOK = true;//false if the message is malformed

    //### Extract the message type ###
    if (receivedBytes.size() > 0)
    {
        messageType = receivedBytes.getByte(0);
        receivedBytes.removeFirst(1);
    }
    else
    {
        reportError("DDispatcher received a UDP message with no type.", true);
        allOK = false;
    }

    //### Verify that we didn't receive the reserved message type ###
    if (allOK)
    {
        if (messageType == RESERVED_MESSAGE_TYPE)
        {
            reportError("DDispatcher received a UDP message with a type reserved for internal (TCP-only) use.", true);
            allOK = false;
        }
    }

    //### Handle the message ###
    if (allOK)
    {
        if (haveSingleHandler)
        {
            udpDataHandler = singleHandler;
            finished = true;
        }
    }

    //### Handle messages with a type but no subtype ###
    if ((allOK) && (!finished))
    {
        msgEventMap = subtypeMaps.at(messageType);

        if (msgEventMap->getHaveSingle())
        {
            udpDataHandler = msgEventMap->getSingleHandler();
            finished = true;
        }
        else if (msgEventMap->size() == 0)
        {
            reportError("No handler for message type " + DEngine::intToStr(messageType), true);
            allOK = false;
        }
    }

    //### Handle messages with a subtype ###
    if ((allOK) && (!finished))
    {
        if (receivedBytes.size() >= 2)
        {
            receivedBytes.extractVariable(0, messageSubType);
            receivedBytes.removeFirst(2);

            if (msgEventMap->getHandler(messageSubType, udpDataHandler))
                finished = true;
            else
            {
                reportError("No subtype set for message type "
                      + DEngine::intToStr(messageType)
                      + " subtype "
                      + DEngine::intToStr(messageSubType), true);
                allOK = false;
            }
        }
        else
        {
            reportError("No subtype set for message type " + DEngine::intToStr(messageType), true);
            allOK = false;
        }
    }

    //### Process the message contents ###
    if (allOK)
        udpDataHandler(receivedBytes);
}//receiveUdpData

DProtocol DDispatcher::generateUdpRequestProtocol()
{
    DProtocol result;//the protocol to return

    result.appendField(new DField("UDP Key", DA_UINT16_T));
    result.appendField(new DField("Requested Port", DA_UINT16_T));
    result.appendField(new DField("Server Port", DA_UINT16_T));

    return result;
}//generateUdpRequestProtocol

bool DDispatcher::requestUdp(DUdpModule *newUdpModule,
                             uint16_t remotePort,
                             boost::function<void (UdpStatus)> newUdpResultHandler)
{
    DField *currentField = NULL;//the current field in udpProtocol to populate
    DProtocol udpProtocol = generateUdpRequestProtocol();
    uint16_t myUdpSendKey = 0;//the key the remote host should use when sending to us
    bool result = true;//false if newUdpModule isn't bound to a port

    //### Remove this DDispatcher from any previously set DUdpModule ###
    disableUdp();

    //### Verify that the new module is bound to a port ###
    if (newUdpModule->getPortNumber() == 0)
        result = false;

    //### Add this DDispatcher to the new DUdpModule ###
    if (result)
    {
        if ((udpModule != newUdpModule) || (remoteUdpPort != remotePort))
        {
            udpModule = newUdpModule;
            myUdpSendKey = udpModule->addDispatcher(this, remotePort, udpRemoveKey);
            remoteUdpPort = remotePort;
        }
    }

    //### Send a UDP Request ###
    if (result)
    {
        currentField = udpProtocol.getField("UDP Key");
        currentField->addValue(myUdpSendKey);

        currentField = udpProtocol.getField("Requested Port");
        currentField->addValue(remotePort);

        currentField = udpProtocol.getField("Server Port");
        currentField->addValue(udpModule->getPortNumber());

        sendCustomTcp(DADD_UDP_REQUEST, udpProtocol.generateMessage());
    }

    //### Wait for the reply ###
    if (result)
    {
        udpResultHandler = newUdpResultHandler;
        waitingForUdpReply = true;
    }

    return result;
}//requestUdp

void DDispatcher::setUdpRequestHandler(boost::function<void (uint16_t, DUdpModule*&)> newUdpRequestHandler)
{
    udpRequestHandler = newUdpRequestHandler;
    haveUdpRequestHandler = true;
}//setUdpRequestHandler

void DDispatcher::disableUdp(void)
{
    if (udpModule != NULL)
    {
        udpModule->removeDispatcher(udpRemoveKey, udpSendKey, this);
        udpModule = NULL;
    }

    udpOK = false;
    udpRemoveKey.clear();
    udpSendKey = 0;
}//disableUDP

void DDispatcher::removeUdpModule(DUdpModule *source)
{
    if (udpModule == NULL)
        DEngine::fatalError("DDispatcher::removeUDPModule(): UDP is not enabled!");

    if (source != udpModule)
        DEngine::fatalError("DDispatcher::removeUDPModule(): DUdpModule does not match!");

    udpOK = false;
    udpModule = NULL;
    udpRemoveKey.clear();
}//removeUdpModule

void DDispatcher::udpRequestEvent(DBuffer &message)
{
    DField *currentField = NULL;//the current field in udpProtocol to read
    DUdpModule *newModule = NULL;//the new UDP module to use
    DProtocol udpProtocol = generateUdpRequestProtocol();
    DBuffer reply;//indicates if we have enabled UDP or not
    uint16_t requestedPort = 0;//the port number the server has asked us to bind to
    uint16_t myUdpSendKey = 0;//the key the remote host should us when sending to us
    bool allOK = true;//false if we shouldn't or can't enable UDP

    //### Check if we have an event handler for the request ###
    if (!haveUdpRequestHandler)
        allOK = false;

    //### Parse the message ###
    if (allOK)
    {
        if (!udpProtocol.parseMessage(message))
        {
            reportError("Received an invalid UDP request: " + getErrMsg(), true);
            allOK = false;
        }
    }

    //###  Call the UDP request handler ###
    if (allOK)
    {
        currentField = udpProtocol.getField("Requested Port");
        currentField->getValue(0, requestedPort);

        udpRequestHandler(requestedPort, newModule);

        if (newModule == NULL)
            allOK = false;
    }//if allOK

    //### Use the new DUdpModule ###
    if (allOK)
    {
        if (newModule->getPortNumber() != requestedPort)
            DEngine::fatalError("DDispatcher::udpRequestEvent(): bound to wrong port!");

        disableUdp();
        udpModule = newModule;

        currentField = udpProtocol.getField("Server Port");
        currentField->getValue(0, remoteUdpPort);

        myUdpSendKey = udpModule->addDispatcher(this, remoteUdpPort, udpRemoveKey);

        currentField = udpProtocol.getField("UDP Key");
        currentField->getValue(0, udpSendKey);

        reply.appendByte(DA_UDP_OK);
        reply.appendVariable(myUdpSendKey);
        udpOK = true;
    }
    else
        reply.appendByte(DA_UDP_REFUSED);

    sendCustomTcp(DADD_UDP_RESULT, reply);
}//udpRequestEvent

void DDispatcher::udpResultEvent(DBuffer &message)
{
    bool allOK = true;//false on malformed message

    //### Verify that we're expecting a UDP reply ###
    if (waitingForUdpReply)
        waitingForUdpReply = false;
    else
    {
        reportError("Received an unexpected UDP reply!", true);
        allOK = false;
    }

    //### Parse the reply ###
    if (allOK)
    {
        //UDP Refused
        if (message.size() == 1)
        {
            if (message.getByte(0) == DA_UDP_REFUSED)
                udpResultHandler(DA_UDP_REFUSED);
            else
            {
                reportError("Received an invalid UDP reply.", true);
                allOK = false;
            }
        }

        //UDP OK
        else if (message.size() == 3)
        {
            if (message.getByte(0) != DA_UDP_OK)
            {
                reportError("Received an invalid UDP reply.", true);
                allOK = false;
            }

            if (allOK)
            {
                message.extractVariable(1, udpSendKey);
                udpOK = true;
                udpResultHandler(DA_UDP_OK);
            }
        }

        //Invalid reply length
        else
        {
            reportError("Received a UDP reply with an invalid length.", true);
            allOK = false;
        }
    }
}//udpResultEvent

bool DDispatcher::udpEnabled(void)
{
    return udpOK;
}//udpEnabled
