/*
 * Copyright (C) 2011-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
  Provides message-based communication. Normally, we must deal with the situation where
  only partial messages are obtained from the TCP stream by waiting until more data arrives.
  DMessageClient seperates the stream into distinct messages, and calls messageHandler
  whenever a complete message arrives. This comes at an overhead of 5 bytes per message,
  2 bytes for the message size, and a 3-byte header that allows us to recover if the TCP stream
  is corrupted.

  Furthermore, the DDispatcher allows us to split a single TCP connection into seperate 'channels'.
  We can add a 1-byte type with a 2-byte subtype, giving us over 16 million possible channels. If no
  type is specified, the DDispatcher won't send any type information over the network, and we keep our
  5 bytes of overhead per message. Likewise, if the subtype isn't specified, we have an overhead of 6 bytes,
  5 from the header and one from the type.
*/

#ifndef DDISPATCHER_HPP_INCLUDED
#define DDISPATCHER_HPP_INCLUDED

#include "DEngine.hpp"
#include "DClient.hpp"
#include "DBuffer.hpp"
#include "DProtocol.hpp"
#include "DField.hpp"

namespace libdasyne
{
    class DEventMap;
    class DFlushEvent;
    class DUdpModule;

    namespace DasyneInt
    {
        //The states in the finite state machine for parsing received messages
        enum DADD_FSMState
        {
            DADD_FIRST_HEADER, DADD_SECOND_HEADER, DADD_THIRD_HEADER, DADD_SIZE_LENGTH,
            DADD_SIZE, DADD_MESSAGE_TYPE, DADD_FIRST_SUBTYPE, DADD_SECOND_SUBTYPE,
            DADD_MESSAGE_BODY
        };//DADD_FSMState

        //The sub types of the custom messages, which have message type RESERVED_MESSAGE_TYPE.
        enum DADD_CustomSubType
        {
            DADD_UDP_REQUEST = 0,   //Called when we receive a UDP request
            DADD_UDP_RESULT = 1     //Called with the result from requestUdp
        };//DADD_CustomSubType
    }//namespace DasyneInt

    enum UdpStatus
    {
        DA_UDP_OK = 0,
        DA_UDP_REFUSED = 1,
        DA_UDP_TIMEOUT = 2
    };//enum UdpStatus

    class DDispatcher : private DClient
    {
        public:
            //constructors
            DDispatcher(DEngine *newNetEngine);
            DDispatcher(DEngine *newNetEngine,
                        unsigned int newBufferSize);

            //constructors, should be called when a DServer accepts a new connection
            DDispatcher(DEngine *newNetEngine,
                        boost::asio::ip::tcp::socket *newSocket);
            DDispatcher(DEngine *newNetEngine,
                        boost::asio::ip::tcp::socket *newSocket,
                        unsigned int newBufferSize);

            //destructor
            virtual ~DDispatcher(void);

            //Functions to expose from DClient
            using DClient::startConnect;
            using DClient::closeSocket;
            using DClient::getErrMsg;
            using DClient::getLocalAddress;
            using DClient::getRemoteAddress;
            using DClient::getBufferSize;
            using DClient::setDelay;
            using DClient::setSendHandler;
            using DClient::setErrorHandler;
            using DClient::unsentBytes;
            using DClient::isConnected;

            //Add an event handler to be called when we receive a message of the
            //specified type
            void addHandler(boost::function<void (DBuffer&)> newHandler);
            void addHandler(uint8_t messageType,
                            boost::function<void (DBuffer&)> newHandler);
            void addHandler(uint8_t messageType,
                            uint16_t subType,
                            boost::function<void (DBuffer&)> newHandler);

            //Removes the specified event handler for received messages.
            void removeHandler(void);
            void removeHandler(uint8_t messageType);
            void removeHandler(uint8_t messageType,
                               uint16_t subType);

            //Sends the specified data to the remote host. If Nagle's algorithm has been disabled,
            //the data is not immediately sent; use flush() to send it. This allows you to
            //add many small messages into a single network packet. If flush is not called, the
            //data will be sent the next time DEngine::update() is called.
            void sendTcp(const libdasyne::DBuffer &theData);
            void sendTcp(uint8_t messageType,
                         const libdasyne::DBuffer &theData);
            void sendTcp(uint8_t messageType,
                         uint16_t subType,
                         const libdasyne::DBuffer &theData);

            //Same as above, but sends the message over UDP. UDP must first be enabled with requestUDP
            //or addUdpRequestHandler. sendHandler is never called for UDP data.
            void sendUdp(const libdasyne::DBuffer &theData);
            void sendUdp(uint8_t messageType,
                         const libdasyne::DBuffer &theData);
            void sendUdp(uint8_t messageType,
                         uint16_t subType,
                         const libdasyne::DBuffer &theData);

            //Enable UDP communcation over the specified socket. The socket should be bound and
            //ready to use. We'll send UDP data to the address returned by getRemoteAddress(), over
            //the remotePort. requestUdp must also be called before any UDP packets can be read.
            //Deleting newUdpModule will automatically call disableUdp() first.
            //Returns false if newUdpModule is not bound to a port
            //Note: calling this function will result in a call to sendHandler, as the request is
            //sent over TCP.
            bool requestUdp(DUdpModule *newUdpModule,
                            uint16_t remotePort,
                            boost::function<void (UdpStatus)> newUdpResultHandler);

            //Specify the handler that will be called when UDP requests are received. The uin16_t is the UDP port number
            //it's asking us to bind to. Set the DUdpModule to NULL if refusing the request.
            void setUdpRequestHandler(boost::function<void (uint16_t, DUdpModule*&)> newUdpRequestHandler);

            //If UDP was enabled by setting the DUdpModule in udpRequestHandler or request UDP,
            //the DDispatcher will stop accepting UDP packets from that module. The module's socket
            //will remain open.
            void disableUdp(void);

            //Send any buffered data from the send() function. If Nagle's algorithm is enabled,
            //calling flush has no effect, as we are still limited by Nagle's. Returns true if there
            //was any data to be flushed, false otherwise.
            bool flush(void);

            //Returns the number of bytes in the TCP send buffer, plus the number of bytes in outMessages
            unsigned long int unsentBytes(void);

            //Returns true if we can currently send and receive UDP messages.
            bool udpEnabled(void);

            //Removes all user-defined handlers
            virtual void clearAllHandlers(void);

            //Three header bytes that are placed at the beginning of each message.
            //These are used to find the next message if we should lose our place in the
            //TCP stream. Assume the longest message size is 64K. Given random binary data,
            //we have even odds of finding any given 2-byte combination within a message.
            //With three bytes, we have only a 1 in 256 chance of finding a false header,
            //and should be able to get back on track with minimal skipped messages.
            static const int FIRST_HEADER_BYTE = 166;
            static const int SECOND_HEADER_BYTE = 9;
            static const int THIRD_HEADER_BYTE = 57;

            //Messages with this type are reserved for internal use only.
            static const int RESERVED_MESSAGE_TYPE = 255;

            //The default message type, used when calling sendTcp without specifying a type.
            static const uint8_t DEFAULT_MESSAGE_TYPE = 0;

            //Allow select classes to access DDispatcher's private methods
            friend class DUdpModule;
            friend class DEngine;

        protected:
            //A description of the most recent error
            using DClient::errMsg;

            //Report that setDelay was called, creates or destroys eventHandler as appropriate
            void delayToggled(void);

        private:
            //Each element is an std::map of event handlers. The element's index is the message type,
            //and the map contains the defined subtypes. If there is no subtype, the DEventMap
            //calls a stored event handler.
            std::vector<DEventMap*> subtypeMaps;

            //Each element is a message that should be sent to the server the next time flush is
            //called.
            std::vector<DBuffer*> outMessages;

            //Calls flush() whenever the DEngine updates. Only used if Nagle's algorithm is disabled
            DFlushEvent *eventHandler;

            //Handles UDP functionality, set to NULL if UDP not enabled
            DUdpModule *udpModule;

            //Event handler called when the requestUDP() function completes
            boost::function<void (UdpStatus)> udpResultHandler;

            //Function to call if there is no type for the received message
            boost::function<void (DBuffer&)> singleHandler;

            //Function to call for the message currently being received
            boost::function<void (DBuffer&)> currentHandler;

            //Event handler to call when we receive a UDP request from the server.
            boost::function<void (uint16_t, DUdpModule*&)> udpRequestHandler;

            //A multi-byte parameter we're extracting from the received message
            DBuffer dataField;

            //The message we're currently receiving
            DBuffer recvMessage;

            //Our UDP receive key, used to remove this DDispatcher from the list of UDP receivers.
            std::string udpRemoveKey;

            //Our current state in the message parsing FSM
            DasyneInt::DADD_FSMState myState;

            //The size of the message we're currently receiving
            uint64_t recvSize;

            //The number of bytes we've received since streamOK became false, used to call errorHandler
            //if we've received more than 64K of data without finding a message header.
            long int junkAmount;

            //The number of DEventMaps in subtypeMaps that contain one or more handlers. Used to keep track
            //of whether or not we have any valid handlers in subtypeMaps.
            int numMsgHandlers;

            //The port to send UDP packets to. This value is only valid if udpModule != NULL.
            uint16_t remoteUdpPort;

            //The two-byte code to add to UDP messages. Only contains a valid value if udpModule
            //is != NULL.
            uint16_t udpSendKey;

            //The type of the message currently being received
            uint8_t recvMsgType;

            //The number of bytes in the message size field
            uint8_t sizeFieldLength;

            //True if we've received a valid message and should call errorHandler if we receive
            //invalid data. Set to false once errorHandler has been called, so that we only call it
            //once per corrupt message.
            bool streamOK;

            //True if we've sent a UDP request, and are awaiting a reply
            bool waitingForUdpReply;

            //True if singleHandler points to a function, false otherwise
            bool haveSingleHandler;

            //True if udpRequestHandler points to a function, false otherwise
            bool haveUdpRequestHandler;

            //True if we've created event handlers for RESERVED_MESSAGE_TYPE messages,
            //false otherwise.
            bool createdCustomHandlers;

            //Indicates that we can send UDP to the remote host, and may receive UDP messages in return.
            bool udpOK;

            //############### FUNCTIONS ###############

            //Returns the protocol used to send Udp requests
            static DProtocol generateUdpRequestProtocol();

            //Initialize member variables, called by the constructors
            void initVars(void);

            //Outputs an error message and throws an exception if messageType is equal
            //to RESERVED_MESSAGE_TYPE. The error message varies based on whether
            //we were sending a message or adding a handler.
            void checkReservedType(uint8_t messageType,
                                   bool isSending);

            //Appends the 3-byte header and the number of bytes in theData to newMessage.
            void appendTcpHeader(DBuffer *newMessage);

            //Appends the size of theData to newMessage
            void appendMessageSize(const DBuffer &theData,
                                   DBuffer *newMessage);

            //Returns true if expected == received, returns false otherwise and calls errorHandler
            bool checkHeaderByte(uint8_t expected,
                                 uint8_t received);

            //Event handler, called when we've received data
            void receiveTcpData(libdasyne::DBuffer &receivedBytes);

            //Deletes every element in outMessages and clears it
            void clearOutMessages(void);

            //Sends newMessage to the remote host
            void dispatch(DBuffer *newMessage);

            //Extracts the message length from the network stream
            void extractMessageLength(uint8_t oneByte);

            //Extract the one-byte message type from the stream
            void extractMessageType(uint8_t oneByte);

            //Throws an exception if we don't have any sub-type handlers defined. This function is called
            //by the removeHandler functions, and takes the message type for debugging purposes.
            void verifySubtypesExist(uint8_t messageType);

            //Used by a DUdpModule to notify its DDispatchers that it's being deleted.
            void removeUdpModule(DUdpModule *source);

            //Event handler called if the remote host sends us a request to enable UDP
            void udpRequestEvent(DBuffer &message);

            //Sends a TCP message with type RESERVED_MESSAGE_TYPE.
            void sendCustomTcp(uint16_t subType,
                               const libdasyne::DBuffer &theData);

            //Create the custom event handlers, listed under RESERVED_MESSAGE_TYPE.
            //Does nothing if they've already been added
            void addCustomHandlers(void);

            //Event handler called with the results of a udpRequest()
            void udpResultEvent(DBuffer &message);

            //Event handler called when we receive UDP data
            void receiveUdpData(DBuffer &receivedBytes);

            //Returns false if we can't send or receive UDP, and generates an error that describes
            //the cause.
            bool verifyUdpIsReady();

    };//DDispatcher
}//namespace libdasyne

#endif
