/*
 * Copyright (C) 2005-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DEngine.hpp"
#include "DBytes.hpp"
#include "DTime.hpp"
#include "DEvent.hpp"
#include "DMutex.hpp"
#include "DDispatcher.hpp"
#include "DSocket.hpp"

using namespace std;
using namespace libdasyne;

EndianType DEngine::endianness = calcEndianness();
string DEngine::fatalMsg;
DMutex DEngine::fatalMutex;
int DEngine::numEngines = 0;

DEngine::DEngine(void)
{
    ioService = new boost::asio::io_service;
    globalTime = new DTime;

    globalTime->updateTime();
    numEngines++;

    pendingEventRemoval = false;
    isFirstUpdate = false;
}//constructor

DEngine::~DEngine(void)
{
    list<DEvent*>::iterator eventIter;//points to an element in eventList
    list<DTime*>::iterator timeIter;//points to an element in timeList

    for (eventIter = eventList.begin(); eventIter != eventList.end(); eventIter++)
        (*eventIter)->reportEngineLost(this);

    for (timeIter = timeList.begin(); timeIter != timeList.end(); timeIter++)
        (*timeIter)->reportEngineLost(this);

    clearPendingDeletions();
    deleteSockets();

    delete ioService;
    ioService = NULL;

    delete globalTime;
    globalTime = NULL;

    numEngines--;
}//destructor

int DEngine::update()
{
    boost::system::system_error *sysError = NULL;//a description of any errors
    size_t numHandlers = 1;//the number of handlers that were called by io_service.run()
    int maxUpdates = MAX_DISPATCH_EVENTS / 2;
    int result = 0;//the number of events we dispatched, or -1 on errors

    //### Update the global time ###
    globalTime->updateTime();

    //### Dispatch custom events ###
    updateIter = eventList.begin();
    while ((updateIter != eventList.end()) && (result < maxUpdates))
    {
        //Fire the current event
        result += (*updateIter)->update();

        //Check if the event removed itself from the event list
        if (pendingEventRemoval)
        {
            pendingEventRemoval = false;
            updateIter = eventList.erase(updateIter);
        }
        else
            updateIter++;

        //Perform any requested deletions
        clearPendingDeletions();
    }

    //### Dispatch ASIO events ###
    try
    {
        if (!isFirstUpdate)
        {
            isFirstUpdate = true;
            ioService->reset();
        }

        while ((numHandlers > 0)  && (result < maxUpdates))
        {
            numHandlers = ioService->poll_one();
            result += numHandlers;
            clearPendingDeletions();
        }

        deleteSockets();
    }
    catch (exception &e)
    {
        sysError = dynamic_cast<boost::system::system_error*>(&e);
        if (sysError == NULL)
        {
            errMsg = "Unknown error: ";
            errMsg.append(e.what());
            result = -1;
        }
        else
        {
            errMsg = "System error: ";
            errMsg.append(e.what());
            result = -1;
        }
    }//catch

    return result;
}//update

void DEngine::clearPendingDeletions()
{
    for (unsigned int i = 0; i < pendingDeletions.size(); i++)
        delete pendingDeletions.at(i);
    pendingDeletions.clear();
}

void DEngine::getCurrentTime(uint32_t &seconds,
                             uint16_t &milliseconds)
{
    seconds = globalTime->getSeconds();
    milliseconds = globalTime->getMilliseconds();
}//getCurrentTime

list<DEvent*>::iterator DEngine::addEvent(DEvent *newEvent)
{
    return eventList.insert(eventList.end(), newEvent);
}//addEvent

void DEngine::removeEvent(list<DEvent*>::iterator eventIter)
{
    if (eventIter == updateIter)
        pendingEventRemoval = true;
    else
        eventList.erase(eventIter);
}//removeEvent

list<DTime*>::iterator DEngine::addTime(DTime *newTime)
{
    return timeList.insert(timeList.end(), newTime);
}//addTime

void DEngine::removeTime(list<DTime*>::iterator timeIter)
{
    timeList.erase(timeIter);
}//removeTime

boost::asio::io_service* DEngine::getIoService(void)
{
    return ioService;
}//getIoService

void DEngine::addFatal(const string &description)
{
    fatalMutex.lock();
    fatalMsg.append(description);
    fatalMutex.unlock();
}//addFatal

void DEngine::fatalError(const string &description)
{
    fatalMutex.lock();

    fatalMsg.append(description);
    cout << fatalMsg << endl;
    fatalMsg.clear();

    fatalMutex.unlock();

    throw 1;
}//fatalError

string DEngine::getFatalErr(void)
{
    string result;//the error message to return

    fatalMutex.lock();

    result = fatalMsg;
    fatalMsg.clear();

    fatalMutex.unlock();

    return result;
}//getFatalErr

string DEngine::getErrMsg(void)
{
    return errMsg;
}//getErrMsg

EndianType DEngine::getEndianness(void)
{
    return endianness;
}//getEndianness

EndianType DEngine::calcEndianness(void)
{
    DBytes testEndian;//test the endianness of this machine
    uint32_t testValue = 1;//should set a single byte in testEndian to 1
    uint8_t oneByte = 0;//a single byte to add to testEndian
    EndianType result = DA_UNKNOWN;//the type of endianness used by this architecture

    //### Add a byte to testEndian ###
    for (unsigned int i = 0; i < sizeof(testValue); i++)
    {
        oneByte = ((unsigned char*)&testValue)[i];
        testEndian.appendByte(oneByte);
    }

    //### Obtain the endianness ###
    if (testEndian.getByte(0) == 1)
        result = DA_LITTLE_ENDIAN;
    else if (testEndian.getByte(3) == 1)
        result = DA_BIG_ENDIAN;
    else
        result = DA_UNKNOWN;

    return result;
}//calcEndianness

bool DEngine::strToInt(const string &text,
                       int &theInt)
{
    stringstream converter;//converts text to an int
    int tempValue = 0;//dont change the parameter if there were errors
    bool result = true;//false on conversion error

    converter << text;
    converter >> tempValue;

    if ((converter.fail()) || (converter.good()))
        result = false;
    else
        theInt = tempValue;

  return result;
}//strToInt

bool DEngine::strToInt(const string &text,
                       long int &theInt)
{
    stringstream converter;//converts text to an int
    long int tempValue = 0;//dont change the parameter if there were errors
    bool result = true;//false on conversion error

    converter << text;
    converter >> tempValue;

    if ((converter.fail()) || (converter.good()))
        result = false;
    else
        theInt = tempValue;

  return result;
}//strToInt

string DEngine::intToStr(long int number)
{
    string result;//string containing converted number
    stringstream converter;//does the conversion

    converter << number;
    converter >> result;

    return result;
}//intToStr

void DEngine::sleep(long int milliseconds)
{
    #ifdef DA_WINDOWS_VERSION
        Sleep(milliseconds);
    #else
        timespec sleepTime;

        if (milliseconds >= 1000)
        {
            sleepTime.tv_sec = milliseconds / 1000;
            milliseconds = milliseconds % 1000;
        }//if milliseconds
        else
            sleepTime.tv_sec = 0;

        sleepTime.tv_nsec = milliseconds * 1000000;

        nanosleep(&sleepTime, NULL);
    #endif
}//sleep

void DEngine::deleteAfter(DObject *object)
{
    pendingDeletions.push_back(object);
}//deleteAfter

void DEngine::deleteAfter(DDispatcher *object)
{
    DObject *dispatcherObject = dynamic_cast<DObject*>(object);

    if (dispatcherObject == NULL)
        fatalError("DEngine::deleteAfter(): couldn't cast to DObject");

    pendingDeletions.push_back(object);
}//deleteAfter

void DEngine::deleteSocket(DSocket *theSocket)
{
    theSocket->closeSocket();
    theSocket->clearAllHandlers();
    pendingSocketDeletions.push_back(theSocket);
}//deleteSocket

void DEngine::deleteSockets()
{
    for (unsigned int i = 0; i < pendingSocketDeletions.size(); i++)
        delete pendingSocketDeletions.at(i);
    pendingSocketDeletions.clear();
}

void DEngine::outputErrorCodes(void)
{
    #ifdef DA_DEBUG_MODE
        cout << boost::system::errc::success << " success SUCCESS" << endl;
        cout << boost::system::errc::address_family_not_supported << " address_family_not_supported EAFNOSUPPORT" << endl;
        cout << boost::system::errc::address_in_use << " address_in_use EADDRINUSE" << endl;
        cout << boost::system::errc::address_not_available << " address_not_available EADDRNOTAVAIL" << endl;
        cout << boost::system::errc::already_connected << " already_connected EISCONN" << endl;
        cout << boost::system::errc::argument_list_too_long << " argument_list_too_long E2BIG" << endl;
        cout << boost::system::errc::argument_out_of_domain << " argument_out_of_domain EDOM" << endl;
        cout << boost::system::errc::bad_address << " bad_address EFAULT" << endl;
        cout << boost::system::errc::bad_file_descriptor << " bad_file_descriptor EBADF" << endl;
        cout << boost::system::errc::bad_message << " bad_message EBADMSG" << endl;
        cout << boost::system::errc::broken_pipe << " broken_pipe EPIPE" << endl;
        cout << boost::system::errc::connection_aborted << " connection_aborted ECONNABORTED" << endl;
        cout << boost::system::errc::connection_already_in_progress << " connection_already_in_progress EALREADY" << endl;
        cout << boost::system::errc::connection_refused << " connection_refused ECONNREFUSED" << endl;
        cout << boost::system::errc::connection_reset << " connection_reset ECONNRESET" << endl;
        cout << boost::system::errc::cross_device_link << " cross_device_link EXDEV" << endl;
        cout << boost::system::errc::destination_address_required << " destination_address_required EDESTADDRREQ" << endl;
        cout << boost::system::errc::device_or_resource_busy << " device_or_resource_busy EBUSY" << endl;
        cout << boost::system::errc::directory_not_empty << " directory_not_empty ENOTEMPTY" << endl;
        cout << boost::system::errc::executable_format_error << " executable_format_error ENOEXEC" << endl;
        cout << boost::system::errc::file_exists << " file_exists EEXIST" << endl;
        cout << boost::system::errc::file_too_large << " file_too_large EFBIG" << endl;
        cout << boost::system::errc::filename_too_long << " filename_too_long ENAMETOOLONG" << endl;
        cout << boost::system::errc::function_not_supported << " function_not_supported ENOSYS" << endl;
        cout << boost::system::errc::host_unreachable << " host_unreachable EHOSTUNREACH" << endl;
        cout << boost::system::errc::identifier_removed << " identifier_removed EIDRM" << endl;
        cout << boost::system::errc::illegal_byte_sequence << " illegal_byte_sequence EILSEQ" << endl;
        cout << boost::system::errc::inappropriate_io_control_operation << " inappropriate_io_control_operation ENOTTY" << endl;
        cout << boost::system::errc::interrupted << " interrupted EINTR" << endl;
        cout << boost::system::errc::invalid_argument << " invalid_argument EINVAL" << endl;
        cout << boost::system::errc::invalid_seek << " invalid_seek ESPIPE" << endl;
        cout << boost::system::errc::io_error << " io_error EIO" << endl;
        cout << boost::system::errc::is_a_directory << " is_a_directory EISDIR" << endl;
        cout << boost::system::errc::message_size << " message_size EMSGSIZE" << endl;
        cout << boost::system::errc::network_down << " network_down ENETDOWN" << endl;
        cout << boost::system::errc::network_reset << " network_reset ENETRESET" << endl;
        cout << boost::system::errc::network_unreachable << " network_unreachable ENETUNREACH" << endl;
        cout << boost::system::errc::no_buffer_space << " no_buffer_space ENOBUFS" << endl;
        cout << boost::system::errc::no_child_process << " no_child_process ECHILD" << endl;
        cout << boost::system::errc::no_link << " no_link ENOLINK" << endl;
        cout << boost::system::errc::no_lock_available << " no_lock_available ENOLCK" << endl;
        cout << boost::system::errc::no_message_available << " no_message_available ENODATA" << endl;
        cout << boost::system::errc::no_message << " no_message ENOMSG" << endl;
        cout << boost::system::errc::no_protocol_option << " no_protocol_option ENOPROTOOPT" << endl;
        cout << boost::system::errc::no_space_on_device << " no_space_on_device ENOSPC" << endl;
        cout << boost::system::errc::no_stream_resources << " no_stream_resources ENOSR" << endl;
        cout << boost::system::errc::no_such_device_or_address << " no_such_device_or_address ENXIO" << endl;
        cout << boost::system::errc::no_such_device << " no_such_device ENODEV" << endl;
        cout << boost::system::errc::no_such_file_or_directory << " no_such_file_or_directory ENOENT" << endl;
        cout << boost::system::errc::no_such_process << " no_such_process ESRCH" << endl;
        cout << boost::system::errc::not_a_directory << " not_a_directory ENOTDIR" << endl;
        cout << boost::system::errc::not_a_socket << " not_a_socket ENOTSOCK" << endl;
        cout << boost::system::errc::not_a_stream << " not_a_stream ENOSTR" << endl;
        cout << boost::system::errc::not_connected << " not_connected ENOTCONN" << endl;
        cout << boost::system::errc::not_enough_memory << " not_enough_memory ENOMEM" << endl;
        cout << boost::system::errc::not_supported << " not_supported ENOTSUP" << endl;
        cout << boost::system::errc::operation_canceled << " operation_canceled ECANCELED" << endl;
        cout << boost::system::errc::operation_in_progress << " operation_in_progress EINPROGRESS" << endl;
        cout << boost::system::errc::operation_not_permitted << " operation_not_permitted EPERM" << endl;
        cout << boost::system::errc::operation_not_supported << " operation_not_supported EOPNOTSUPP" << endl;
        cout << boost::system::errc::operation_would_block << " operation_would_block EWOULDBLOCK" << endl;
        cout << boost::system::errc::owner_dead << " owner_dead EOWNERDEAD" << endl;
        cout << boost::system::errc::permission_denied << " permission_denied EACCES" << endl;
        cout << boost::system::errc::protocol_error << " protocol_error EPROTO" << endl;
        cout << boost::system::errc::protocol_not_supported << " protocol_not_supported EPROTONOSUPPORT" << endl;
        cout << boost::system::errc::read_only_file_system << " read_only_file_system EROFS" << endl;
        cout << boost::system::errc::resource_deadlock_would_occur << " resource_deadlock_would_occur EDEADLK" << endl;
        cout << boost::system::errc::resource_unavailable_try_again << " resource_unavailable_try_again EAGAIN" << endl;
        cout << boost::system::errc::result_out_of_range << " result_out_of_range ERANGE" << endl;
        cout << boost::system::errc::state_not_recoverable << " state_not recoverable ENOTRECOVERABLE" << endl;
        cout << boost::system::errc::stream_timeout << " stream_timeout ETIME" << endl;
        cout << boost::system::errc::text_file_busy << " text_file_busy ETXTBSY" << endl;
        cout << boost::system::errc::timed_out << " timed_out ETIMEDOUT" << endl;
        cout << boost::system::errc::too_many_files_open_in_system << " too_many_files_open_in_system ENFILE" << endl;
        cout << boost::system::errc::too_many_files_open << " too_many_files_open EMFILE" << endl;
        cout << boost::system::errc::too_many_links << " too_many_links EMLINK" << endl;
        //cout << boost::system::errc::too_many_synbolic_link_levels << " too_many_symbolic_links ELOOP" << endl;
        cout << boost::system::errc::value_too_large << " value_too_large EOVERFLOW" << endl;
        cout << boost::system::errc::wrong_protocol_type << " wrong_protocol_type EPROTOTYPE" << endl;
    #endif
}//outputErrorCodes
