/*
 * Copyright (C) 2012-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DField.hpp"
#include "DProtocol.hpp"

using namespace std;
using namespace libdasyne;

DField::DField(const std::string &newName,
               FieldType newType)
{
    myType = newType;
    myName = newName;

    minEntries = 1;
    maxEntries = 1;

    //### Assign a generator to create objects of the desired type ###
    switch (myType)
    {
        case  DA_UINT8_T:
            generator = new DField_UINT8_T;
        break;

        case DA_UINT16_T:
            generator = new DField_UINT16_T;
        break;

        case DA_UINT32_T:
            generator = new DField_UINT32_T;
        break;

        case DA_UINT64_T:
            generator = new DField_UINT64_T;
        break;

        case DA_INT8_T:
            generator = new DField_INT8_T;
        break;

        case DA_INT16_T:
            generator = new DField_INT16_T;
        break;

        case DA_INT32_T:
            generator = new DField_INT32_T;
        break;

        case DA_INT64_T:
            generator = new DField_INT64_T;
        break;

        case DA_FLOAT:
            generator = new DField_FLOAT;
        break;

        case DA_DOUBLE:
            generator = new DField_DOUBLE;
        break;

        case DA_STRING:
            generator = new DField_STRING;
        break;

        case DA_DBUFFER:
            generator = new DField_DBUFFER;
        break;

        default:
            DEngine::fatalError("DField::DField(): unknown parameter type");
    }//switch myType
}//constructor

DField::DField(const DField &source)
{
    copyFrom(source);
}//copy constructor

DField::~DField(void)
{
    clear();

    delete generator;
    generator = NULL;
}//destructor

DField& DField::operator= (const DField &source)
{
    if (this != &source)
    {
        clear();

        delete generator;
        generator = NULL;

        copyFrom(source);
    }//if this

    return *this;
}//=

void DField::copyFrom(const DField &source)
{
    DFieldValue *currentValue = NULL;//the current field value to copy
    DBuffer storedBytes;//the value stored in the current DFieldValue

    //### Copy the DField generator ###
    generator = source.generator->generate();
    storedBytes = source.generator->getValue();
    generator->setValue(storedBytes);

    //### Copy theData ###
    for (unsigned int i = 0; i < source.theData.size(); i++)
    {
        currentValue = generator->generate();
        storedBytes = source.theData.at(i)->getValue();

        currentValue->setValue(storedBytes);
        theData.push_back(currentValue);
    }//for i

    //### Copy simple member variables ###
    myName = source.myName;
    myType = source.myType;
    maxEntries = source.maxEntries;
    minEntries = source.minEntries;
}//copyFrom

void DField::generateMessage(DBuffer &theMessage)
{
    DFieldValue *currentField = NULL;//the current field to send
    uint64_t fieldSize = 0;//the size of the current field in bytes

    //### Add the number of values ###
    if ((minEntries != 1) && (maxEntries != 1))
        DProtocol::appendSize(theMessage, theData.size());

    //### Add values with varying lengths ###
    if ((myType == DA_STRING) || (myType == DA_DBUFFER))
    {
        for (unsigned int i = 0; i < theData.size(); i++)
        {
            currentField = theData.at(i);
            fieldSize = currentField->getValue().size();

            DProtocol::appendSize(theMessage, fieldSize);
            theMessage.append(currentField->getValue());
        }//for i
    }//if myType || myType

    //### Add values with constant lengths ###
    else
    {
        for (unsigned int i = 0; i < theData.size(); i++)
            theMessage.append(theData.at(i)->getValue());
    }//else myType || myType
}//generateMessage

void DField::addString(const std::string &newValue)
{
    DField_STRING *newEntry = NULL;//The new item to add to theData

    if (myType != DA_STRING)
        DEngine::fatalError("DField::addString(): this DField does not contain strings");

    if (theData.size() >= maxEntries)
        DEngine::fatalError("DField::addString(): would exceed the max number of entries for this field.");

    newEntry = new DField_STRING;
    newEntry->theValue = newValue;
    theData.push_back(newEntry);
}//addString

void DField::addDBuffer(DBuffer &newValue)
{
    DFieldValue *newEntry = NULL;//The new item to add to theData

    if (myType != DA_DBUFFER)
        DEngine::fatalError("DField::addDBuffer(): this DField does not contain DBuffers");

    if (theData.size() >= maxEntries)
        DEngine::fatalError("DField::addDBuffer(): would exceed the max number of entries for this field.");

    newEntry = generator->generate();
    newEntry->setValue(newValue);
    theData.push_back(newEntry);
}//addDBuffer

void DField::getString(unsigned int index,
                       std::string &theValue) const
{
    DBuffer valueBytes = theData.at(index)->getValue();

    if (myType != DA_STRING)
        DEngine::fatalError("DField::getString(): this DField does not contain strings");

    theValue = valueBytes.toString();
}//getString

void DField::getDBuffer(unsigned int index,
                        DBuffer &theValue) const
{
    if (myType != DA_DBUFFER)
        DEngine::fatalError("DField::getString(): this DField does not contain DBuffers");

    theValue = theData.at(index)->getValue();
}//getDBuffer

void DField::setFieldRange(uint64_t newMinEntries,
                           uint64_t newMaxEntries)
{
    minEntries = newMinEntries;
    maxEntries = newMaxEntries;

    //### Verify that maxEntries is >= minEntries ###
    if (minEntries > maxEntries)
        DEngine::fatalError("DField::setNumValues(): maxEntries < minEntries");

    //### Verify that the vector can contain at least one item ###
    if (maxEntries == 0)
        DEngine::fatalError("DField::setNumValues(): maxEntries is zero");
}//setFieldRange

bool DField::parse(const DBuffer &theMessage,
                   unsigned int &byteIndex,
                   string &errorDesc)
{
    DFieldValue *newValue = NULL;//the new value to add to theData
    DBuffer currentSegment;//The current segment of the message to extract
    uint64_t numValues = 0;//the number of values being sent to this field
    uint64_t valueLength = 0;//the number of bytes in the current value
    bool result = true;//false on malformed entry

    //### Extract the number of values ###
    if ((minEntries != 1) && (maxEntries != 1))
    {
        if (!DProtocol::extractLength(theMessage, byteIndex, numValues, errorDesc))
        {
            addFieldDetails(byteIndex, errorDesc);
            result = false;
        }//if !extractLength

        //Verify that the number of values is valid
        if (result)
        {
            if ((numValues < minEntries) || (numValues > maxEntries))
            {
                errorDesc = "Received ";
                errorDesc.append(DEngine::intToStr(numValues));
                errorDesc.append(" values, but field takes between ");
                errorDesc.append(DEngine::intToStr(minEntries));
                errorDesc.append(" and " );
                errorDesc.append(DEngine::intToStr(maxEntries));
                errorDesc.append(" values");
                addFieldDetails(byteIndex, errorDesc);

                result = false;
            }//if numValues || numValues
        }//if result
    }//if minEntries && maxEntries
    else
        numValues = 1;

    //### Read values with a variable length ###
    if ((myType == DA_STRING) || (myType == DA_DBUFFER))
    {
        while ((result) && (theData.size() < numValues))
        {
            if (!DProtocol::extractLength(theMessage, byteIndex, valueLength, errorDesc))
            {
                addFieldDetails(byteIndex, errorDesc);
                result = false;
            }//if !extractlength()

            if (result)
            {
                if (DProtocol::getNextBytes(theMessage, byteIndex, currentSegment, valueLength))
                {
                    newValue = generator->generate();
                    newValue->setValue(currentSegment);
                    theData.push_back(newValue);
                }//if getNextBytes()

                else
                {
                    errorDesc = "Expected ";
                    errorDesc.append(DEngine::intToStr(numValues));
                    errorDesc.append(" values, but message ended after ");
                    errorDesc.append(DEngine::intToStr(theData.size()));
                    addFieldDetails(byteIndex, errorDesc);

                    result = false;
                }//else getNextBytes()
            }//if result
        }//while result && size()
    }//if myType || myType

    //### Read values with a fixed length ###
    else
    {
        while ((result) && (theData.size() < numValues))
        {
            if (!DProtocol::getNextBytes(theMessage, byteIndex, currentSegment, generator->size()))
            {
                errorDesc = "Expected ";
                errorDesc.append(DEngine::intToStr(numValues));
                errorDesc.append(" values, but message ended after ");
                errorDesc.append(DEngine::intToStr(theData.size()));
                addFieldDetails(byteIndex, errorDesc);

                result = false;
            }//if !getNextBytes()

            if (result)
            {
                newValue = generator->generate();
                newValue->setValue(currentSegment);
                theData.push_back(newValue);
            }//if result
        }//while result && valueIndex
    }//else myType || myType

    return result;
}//parse

void DField::addFieldDetails(unsigned int byteIndex,
                             string &errorDesc)
{
    errorDesc.append(" Error in field ");
    errorDesc.append(myName);
    errorDesc.append(" at byte number ");
    errorDesc.append(DEngine::intToStr(byteIndex + 1));
}//addFieldDetails

void DField::clear(void)
{
    for (unsigned int i = 0; i < theData.size(); i++)
        delete theData.at(i);
    theData.clear();
}//clear

uint64_t DField::size(void)
{
    return theData.size();
}//size

std::string DField::name(void)
{
    return myName;
}//name
