/*
 * Copyright (C) 2012-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    Contains the value from a single field element in a DProtocol.
    That is to say, the DProtocol holds an std::map of DFields.
    DFields hold a vector of DFieldValues.
    And the DFieldValue contains the actual value.

    Ideally, this class would be implemented with a class template
    like this: template <class T>, and would have a member variable of type T.
    However, this has proved to be impossible - any class with a type 'T' member variable
    must have the corresponding class template. Which means we can never have two
    template types in the same std::map, no matter how we subclass with wrappers. If
    DFieldValue is of type T, then DField must also be of type T, as well as DProtocol.

    To get around this limitation, we can store the field's bytes in a DBuffer, and
    re-construct the field with the appropriate sub-class.
*/

#ifndef DFIELDVALUE_HPP_INCLUDED
#define DFIELDVALUE_HPP_INCLUDED

#include "DEngine.hpp"
#include "DBuffer.hpp"

namespace libdasyne
{
    class DFieldValue : public DObject
    {
        public:
            //constructor
            DFieldValue(void);

            //destructor
            virtual ~DFieldValue(void);

            //Generates a new object with the child's type
            virtual DFieldValue* generate(void) = 0;

            //Mutator. Sets the stored value from the bytes in newValue.
            virtual void setValue(DBuffer &newValue) = 0;

            //Accessor. The DBuffer stores the variable contents.
            virtual DBuffer getValue(void) = 0;

            //Returns the number of bytes in this value. Should be called for
            //DFieldValues that contain primitives only.
            virtual uint8_t size() = 0;
    };//DFieldValue

    class DField_UINT8_T : public DFieldValue
    {
        public:
            DField_UINT8_T* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            uint8_t theValue;
    };//class DField_UINT8_T

    class DField_UINT16_T : public DFieldValue
    {
        public:
            DField_UINT16_T* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            uint16_t theValue;
    };//class DField_UINT16_T

    class DField_UINT32_T : public DFieldValue
    {
        public:
            DField_UINT32_T* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            uint32_t theValue;
    };//class DField_UINT32_T

    class DField_UINT64_T : public DFieldValue
    {
        public:
            DField_UINT64_T* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            uint64_t theValue;
    };//class DField_UINT64_T

    class DField_INT8_T : public DFieldValue
    {
        public:
            DField_INT8_T* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            int8_t theValue;
    };//class DField_INT8_T

    class DField_INT16_T : public DFieldValue
    {
        public:
            DField_INT16_T* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            int16_t theValue;
    };//class DField_INT16_T

    class DField_INT32_T : public DFieldValue
    {
        public:
            DField_INT32_T* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            int32_t theValue;
    };//class DField_INT32_T

    class DField_INT64_T : public DFieldValue
    {
        public:
            DField_INT64_T* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            int64_t theValue;
    };//class DField_INT64_T

    class DField_FLOAT : public DFieldValue
    {
        public:
            DField_FLOAT* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            float theValue;
    };//class DField_FLOAT

    class DField_DOUBLE : public DFieldValue
    {
        public:
            DField_DOUBLE* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            double theValue;
    };//class DField_DOUBLE

    class DField_STRING : public DFieldValue
    {
        public:
            DField_STRING* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            std::string theValue;
    };//class DField_STRING

    class DField_DBUFFER : public DFieldValue
    {
        public:
            DField_DBUFFER* generate(void);
            void setValue(DBuffer &newValue);
            DBuffer getValue(void);
            uint8_t size(void);

            DBuffer theValue;
    };//DField_DBUFFER
}//namespace

#endif
