/*
 * Copyright (C) 2011-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    Mimics an array, but performs bounds-checking when an element is accessed
    with []. Used to debug DBytes, and check for buffer overruns. Because it's
    a drop-in replacement for an array, the copy constructor and = operator
    perform shallow copies, which is what an array would do.

*/

#ifndef DSAFEARRAY_HPP_INCLUDED
#define DSAFEARRAY_HPP_INCLUDED

#include "DEngine.hpp"

namespace libdasyne
{
    class DSafeArray : public DObject
    {
        public:
            //constructor
            DSafeArray(long int newSize,
                       bool newIsTemporary);

            //Copy constructor, performs a shallow copy
            DSafeArray(const DSafeArray &source);

            //destructor
            ~DSafeArray(void);

            //If this is a temporary object, create the vector of bytes and make
            //it permanent.
            void makePermanent(void);

            //Access an element in theBytes
            uint8_t& operator [] (long int index);

            //Performs a shallow copy
            DSafeArray& operator = (const DSafeArray &source);

            //Returns the number of DSafeArrays in existence
            static long int getNumInstances(void);

        private:
            //The data being stored
            std::vector<uint8_t> *theBytes;

            //The number of DSafeArrays in existence
            static long int numInstances;

            //The size of theBytes, should be used by objects copying from this one
            unsigned int temporarySize;

            //If true, don't delete theBytes when the destructor is called
            bool isShallowCopy;

            //If true, this is a temporary variable that's about to be used as source
            //in another instance's copy constructor. Don't create theBytes, instead
            //set temporarySize to the number of bytes that would have been allocated.
            bool isTemporary;

    };//DSafeArray
}//namespace libdasyne

#endif
