/*
 * Copyright (C) 2005-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DTime.hpp"
#include "DEngine.hpp"

using namespace std;
using namespace libdasyne;

DTime::DTime(void)
{
    createWinTimer();

    seconds = 0;
    milliseconds = 0;
    engine = NULL;
}//constructor

DTime::DTime(const DTime &source)
{
    seconds = source.seconds;
    milliseconds = source.milliseconds;
    useEngine(source.engine);

    createWinTimer();
}//copy constructor

DTime::~DTime(void)
{
    leaveEngine();
}//destructor

DTime& DTime::operator= (const DTime &source)
{
    if (this != &source)
    {
        seconds = source.seconds;
        milliseconds = source.milliseconds;
        useEngine(source.engine);
    }//if this

    return *this;
}//=

void DTime::createWinTimer(void)
{
    #ifdef DA_WINDOWS_VERSION
        LARGE_INTEGER performanceFreq;//extracts the frequency of the high res timer

        performanceFreq.QuadPart = 0;
        winTimerValue.QuadPart = 0;
        winCounterDiv = 0;

        if (QueryPerformanceFrequency(&performanceFreq))
        {
            winCounterDiv = performanceFreq.QuadPart;
            if (winCounterDiv != 0)
            {
                winCounterDiv = winCounterDiv / 1000;
                if (winCounterDiv == 0)
                    DEngine::fatalError("DTime::createWinTimer(): high performance counter is updating less than 1000 x per sec");
            }//if winCounterDiv
            else
                DEngine::fatalError("DTime::createWinTimer(): high performance counter is updating zero times per second!");
        }//if QueryPerformanceFrequency()
        else
            DEngine::fatalError("DTime::createWinTimer(): QueryPerformanceFrequency() failed!");
    #endif
}//createWinTimer

void DTime::useEngine(DEngine *newEngine)
{
    leaveEngine();

    if (newEngine != NULL)
    {
        engine = newEngine;
        myIter = engine->addTime(this);
    }//if newEngine
}//useEngine

void DTime::leaveEngine(void)
{
    if (engine != NULL)
    {
        engine->removeTime(myIter);
        engine = NULL;
    }//if engine
}//leaveEngine

void DTime::reportEngineLost(DEngine *theEngine)
{
    if (engine != theEngine)
        DEngine::fatalError("DTime::reportEngineLost(): that's not my engine!");

    engine = NULL;
}//reportEngineLost

void DTime::updateTime(void)
{
    if (engine == NULL)
        getCurrentTime(seconds, milliseconds);
    else
        engine->getCurrentTime(seconds, milliseconds);
}//updateTime

uint32_t DTime::timeElapsed(void)
{
    uint32_t result = 0;//the number of elapsed milliseconds to return
    uint32_t currentSeconds = 0;//current time, in seconds
    long int timeAmount = 0;//check for negative time elapsed values
    uint16_t currentMilliseconds = 0;//millisecond component of the current time

    //### Get number of elapsed seconds ###
    if (engine == NULL)
        getCurrentTime(currentSeconds, currentMilliseconds);
    else
        engine->getCurrentTime(currentSeconds, currentMilliseconds);
    timeAmount = currentSeconds - seconds;

    //### Make sure the number of milliseconds will fit in a 32-bit signed variable ###
    if (timeAmount > 2147482)
        timeAmount = 2147482;

    //### Convert to milliseconds ###
    if (timeAmount >= 0)
    {
        timeAmount *= 1000;
        timeAmount += currentMilliseconds - milliseconds;

        if (timeAmount >= 0)
            result = timeAmount;
    }//if timeAmount

    return result;
}//timeElapsed

void DTime::getCurrentTime(uint32_t &currentSeconds,
                            uint16_t &currentMilliseconds)
{
    #ifdef DA_WINDOWS_VERSION
        if (QueryPerformanceCounter(&winTimerValue))
        {
            if (winTimerValue.QuadPart != 0)
                winTimerValue.QuadPart = winTimerValue.QuadPart / winCounterDiv;

            if (winTimerValue.QuadPart != 0)
            {
                currentSeconds = winTimerValue.QuadPart / 1000;
                currentMilliseconds = winTimerValue.QuadPart % 1000;
            }//if QuadPart
        }//if QueryPerformanceCounter()
        else
            DEngine::fatalError("DTime::updateTime(): Couldn't access the high res timer.");

    #else
        timeval currentTime;//the current time, microsecond accuracy
        uint32_t microseconds = 0;//used to store the time in microseconds

        if (gettimeofday(&currentTime, NULL) == 0)
        {
            currentSeconds = currentTime.tv_sec;
            microseconds = currentTime.tv_usec;
            if (microseconds != 0)
                microseconds = microseconds / 1000;
            currentMilliseconds = microseconds;
        }//if gettimeofday()
        else
            DEngine::fatalError("DTime::updateTime(): gettimeofday() failed");
    #endif
}//getCurrentTime

uint32_t DTime::getSeconds(void)
{
    return seconds;
}//getSeconds

uint16_t DTime::getMilliseconds(void)
{
    return milliseconds;
}//getMilliseconds

void DTime::addTime(uint32_t newTime)
{
    newTime += milliseconds;

    if (newTime != 0)
    {
        seconds += newTime / 1000;
        milliseconds = newTime % 1000;
    }//if newTime
}//addTime

void DTime::setTime(uint32_t newSeconds,
                    uint16_t newMilliseconds)
{
    seconds = newSeconds;
    milliseconds = newMilliseconds;
}//setTime
