/*
 * Copyright (C) 2005-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
  A simple millisecond-accurate timer. Each DTime object contains a stored time,
  which can be compared to the current time to determine the time elapsed.
*/

#ifndef DTIME_HPP_INCLUDED
#define DTIME_HPP_INCLUDED

#include "DEngine.hpp"

namespace libdasyne
{
    class DTime : public DObject
    {
        public:
            //constructor
            DTime(void);

            //copy constructor
            DTime(const DTime &source);

            //destructor
            ~DTime(void);

            //Overloaded = operator
            DTime& operator= (const DTime &source);

            //updates the stored time to the current time
            void updateTime(void);

            //returns the number of milliseconds since the last time updateTime() was called
            uint32_t timeElapsed(void);

            //increases the stored time by the specified number of milliseconds
            void addTime(uint32_t newTime);

            //Sets the stored time to the specified values
            void setTime(uint32_t newSeconds,
                         uint16_t newMilliseconds);

            //return the stored time
            uint32_t getSeconds(void);
            uint16_t getMilliseconds(void);

            //The DEngine will set it's current time whenever DEngine::update() is called.
            //Calling updateTime() will return the DEngine's time, rather than querying the operating
            //system directly. This can improve performance when many timers are being used, since
            //we avoid expensive system calls. However, it reduces the resolution of the DTime to the number
            //of times DEngine::update is called per second.
            //Set newEngine to NULL to query the OS time directly.
            void useEngine(DEngine *newEngine);

            //Report that the DEngine has been deleted
            void reportEngineLost(DEngine *theEngine);

        private:
            #ifdef DA_WINDOWS_VERSION
                //current value of the high-res timer
                LARGE_INTEGER winTimerValue;

                //number of high-res timer updates/sec, divided by 1000. Divide
                //the current winTimerValue by this to get milliseconds
                uint64_t winCounterDiv;
            #endif

            //Pointer to the DEngine to use for timing values, or NULL if we'll obtain them ourselves
            DEngine *engine;

            //Iterator to my location in the engine's time list. Only valid if engine != NULL
            std::list<DTime*>::iterator myIter;

            //the time being stored
            uint32_t seconds;
            uint16_t milliseconds;

            //############### FUNCTIONS ###############

            //Creates the windows high-res timer
            void createWinTimer(void);

            //Obtains the current time, stores it in the provided variables
            void getCurrentTime(uint32_t &currentSeconds,
                                uint16_t &currentMilliseconds);

            //De-registers with the current engine, if any.
            void leaveEngine(void);

    };//class DTime
}//namespace libdasyne
#endif
