/*
 * Copyright (C) 2014-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DUdpModule.hpp"
#include "DDispatcher.hpp"

using namespace libdasyne;
using namespace std;

DUdpModule::DUdpModule(DEngine *newNetEngine) : DUdpSocket(newNetEngine)
{
    prevKey = 0;
    keyStamper = 0;
}//constructor

DUdpModule::~DUdpModule(void)
{
    map<string, DDispatcher*>::iterator udpIter;//the element in udpReceivers to notify

    for (udpIter = udpReceivers.begin(); udpIter != udpReceivers.end(); udpIter++)
        udpIter->second->removeUdpModule(this);

    udpReceivers.clear();
}//destructor

bool DUdpModule::closeSocket(void)
{
    return DUdpSocket::closeSocket();
}//closeSocket

bool DUdpModule::bind(IpVersion ipVersion,
                      uint16_t portNumber)
{
    setReceiveHandler(boost::bind(&DUdpModule::moduleReceiver, this, _1, _2, _3));

    return DUdpSocket::bind(ipVersion, portNumber);
}//bind

void DUdpModule::moduleReceiver(DBuffer &theMessage,
                                string &remoteAddress,
                                uint16_t remotePort)
{
    DDispatcher *receiver = NULL;//the DDispatcher to pass this message to
    map<string, DDispatcher*>::iterator udpIter;//the element in udpReceivers to receive this message
    string mapKey;//the DDispatcher's key in the map
    uint16_t udpKey = 0;//the receiver's UDP key
    bool allOK = true;//false on errors

    //### Extract the UDP key ###
    if (theMessage.size() >= 2)
    {
        theMessage.extractVariable(0, udpKey);
        theMessage.removeFirst(2);
    }
    else
    {
        reportError("Received a UDP packet without a key", true);
        allOK = false;
    }

    //### Find the receiving DDispatcher ###
    if (allOK)
    {
        mapKey = generateMapKey(udpKey, remotePort, remoteAddress);
        udpIter = udpReceivers.find(mapKey);

        if (udpIter != udpReceivers.end())
            receiver = udpIter->second;
        else
        {
            reportError("Received a UDP packet with an unknown key", true);
            allOK = false;
        }
    }

    //### Pass the message to the DDispatcher ###
    if (allOK)
        receiver->receiveUdpData(theMessage);
}//moduleReceiver

string DUdpModule::generateMapKey(uint16_t udpSendKey,
                                  uint16_t remotePort,
                                  string remoteAddress)
{
    return DEngine::intToStr(udpSendKey)
         + "|" + DEngine::intToStr(remotePort)
         + "|" + remoteAddress;
}//generateMapKey

void DUdpModule::removeDispatcher(string mapKey,
                                  uint16_t udpSendKey,
                                  DDispatcher *theDispatcher)
{
    map<string, DDispatcher*>::iterator udpIter;//the element in udpReceivers to remove

    //### Find the position of the DDispatcher to remove ###
    udpIter = udpReceivers.find(mapKey);
    if (udpIter == udpReceivers.end())
    {
        DEngine::addFatal("DUdpModule::removeDispatcher(): key not found: ");
        DEngine::fatalError(mapKey);
    }

    //### Verify that this is the proper DDispatcher ###
    if (udpIter->second == theDispatcher)
        udpIter->second = NULL;
    else
        DEngine::fatalError("DUdpModule::removeDispatcher(): DDispatchers do not match!");

    //### Perform the removal ###
    availableKeys.push(udpSendKey);
    udpReceivers.erase(udpIter);
}//removeDispatcher

uint16_t DUdpModule::addDispatcher(DDispatcher *theDispatcher,
                                   uint16_t remoteUdpPort,
                                   string &mapKey)
{
    pair<map<string, DDispatcher*>::iterator, bool> insertReturn;//the return value from the insert operation
    uint16_t result = 0;//2-byte key that will be sent in UDP packets to signify the DDispatcher to call

    //### Find an available send key ###
    if (availableKeys.size() == 0)
    {
        if (keyStamper < 65535)
        {
            availableKeys.push(keyStamper);
            keyStamper++;
        }
        else
            DEngine::fatalError("DUdpModule::addDispatcher(): exceeded maximum possible number of connections: 65536");
    }
    result = availableKeys.front();
    availableKeys.pop();

    //### Insert the DDispatcher into udpReceivers ###
    mapKey = generateMapKey(result, remoteUdpPort, theDispatcher->getRemoteAddress());
    insertReturn = udpReceivers.insert(pair<string, DDispatcher*>(mapKey, theDispatcher));

    if (!insertReturn.second)
    {
        DEngine::addFatal("DUdpModule::addDispatcher(): duplicate map key ");
        DEngine::fatalError(DEngine::intToStr(result));
    }

    return result;
}//addDispatcher
