/*
 * Copyright (C) 2014-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    Provides UDP communication to DDispatchers. We construct a key for each incoming
    datagram, consisting of the datagram's source address and port number, as well
    as a two-byte value contained within the datagram payload.

    The key is used to find the DDispatcher that should receive the datagram

    DUdpModule inherits from DUpdSocket privately so that we can hide
    the setSendHandler and setReceiveHandler methods, which make no sense to use here.
*/

#ifndef DUDPMODULE_HPP_INCLUDED
#define DUDPMODULE_HPP_INCLUDED

#include "DUdpSocket.hpp"
#include "DEngine.hpp"

#include <queue>

namespace libdasyne
{
    class DDispatcher;

    class DUdpModule : private DUdpSocket
    {
        public:
            //constructor
            DUdpModule(DEngine *newNetEngine);

            //destructor
            ~DUdpModule(void);

            //Functions to expose from DUdpSocket
            using DUdpSocket::resolveHost;
            using DUdpSocket::setErrorHandler;
            using DUdpSocket::getErrMsg;
            using DUdpSocket::getPortNumber;

            //Bind the UDP socket to the specified port number. Returns false if an error occurred
            //and sets errMsg.
            bool bind(IpVersion ipVersion,
                      uint16_t portNumber);

            //Closes the socket, notifying any DDispatchers that are using this
            //DUdpModule. Returns false on errors and sets errMsg.
            bool closeSocket(void);

            //Allow DDispatcher to access our private methods and members
            friend class DDispatcher;

        private:
            //The key is a unique 2-byte code corresponding to a TCP connection,
            //plus the remote host's IP address and port number
            std::map<std::string, DDispatcher*> udpReceivers;

            //Contains available 2-byte keys. To keep memory requirements small, we only add keys
            //to it as needed, up to a maximum of 65536
            std::queue<uint16_t> availableKeys;

            //The current key to add to available keys. It's incremented by 1 after being added.
            uint32_t keyStamper;

            //The previously used key, for adding a DDispatcher to udpReceivers
            int prevKey;

            //############### Functions ###############

            //Removes the specified DDispatcher from our event handlers.
            void removeDispatcher(std::string removeKey,
                                  uint16_t mapKey,
                                  DDispatcher *theDispatcher);

            //Add a DDispatcher to the map of UDP receivers. Returns the send key,
            //and sets mapKey to the DDispatcher's position in the map, used to remove it.
            uint16_t addDispatcher(DDispatcher *theDispatcher,
                                   uint16_t remoteUdpPort,
                                   std::string &mapKey);

            //Receives all UDP messages and sends them to the proper DDispatcher
            void moduleReceiver(DBuffer &theMessage,
                                std::string &remoteAddress,
                                uint16_t remotePort);

            //Given a udp send key, remote port and address, generate the map key to refer to
            //a DDispatcher in udpReceivers.
            std::string generateMapKey(uint16_t udpSendKey,
                                       uint16_t remotePort,
                                       std::string remoteAddress);

    };//DUdpModule
}//namespace libdasyne

#endif
