/*
 * Copyright (C) 2014-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

#include "DUdpSocket.hpp"
#include "DBuffer.hpp"
#include "DUdpMessage.hpp"

using namespace std;
using namespace libdasyne;

DUdpSocket::DUdpSocket(DEngine *newNetEngine)
{
    netEngine = newNetEngine;

    myPort = 0;
    isSending = false;
    socketBound = false;
    clearAllHandlers();

    udpSocket = new boost::asio::ip::udp::socket(*netEngine->getIoService());
    hostResolver = new boost::asio::ip::udp::resolver(*netEngine->getIoService());
    receiveBuffer = new unsigned char[RECEIVE_BUFFER_SIZE];
}//constructor

DUdpSocket::~DUdpSocket(void)
{
    closeSocket();

    delete udpSocket;
    udpSocket = NULL;

    delete hostResolver;
    hostResolver = NULL;

    delete [] receiveBuffer;
    receiveBuffer = NULL;

    for (unsigned int i = 0; i < sendQueue.size(); i++)
        delete sendQueue.at(i);
    sendQueue.clear();
}//destructor

void DUdpSocket::resolveHost(const std::string &hostName,
                             libdasyne::IpVersion ipVersion,
                             boost::function<void (vector<string>)> newResolveHandler)
{
    boost::asio::ip::udp::resolver::query *nameQuery = NULL;//query to resolve hostName to an IP address

    //### Store the resolveHandler function ###
    if (haveResolveHandler)
        DEngine::fatalError("DUdpSocket::resolveHost(): a host name resolution is already in progress");
    else
    {
        resolveHandler = newResolveHandler;
        haveResolveHandler = true;
    }//else haveResolveHandler

    //### Create a query for this host name ###
    if (ipVersion == DA_IPV4)
        nameQuery = new boost::asio::ip::udp::resolver::query(boost::asio::ip::udp::v4(), hostName, "",
                                                                boost::asio::ip::resolver_query_base::address_configured);
    else if (ipVersion == DA_IPV6)
        nameQuery = new boost::asio::ip::udp::resolver::query(boost::asio::ip::udp::v6(), hostName, "",
                                                                boost::asio::ip::resolver_query_base::address_configured);

    //resolver_query_base::flags resolve_flags = address_configured

    //### Perform the resolution ###
    hostResolver->async_resolve(*nameQuery, boost::bind(&DUdpSocket::resolveFinished, this,
                                                        boost::asio::placeholders::error,
                                                        boost::asio::placeholders::iterator));

    //### Free memory ###
    delete nameQuery;
    nameQuery = NULL;
}//resolveHost

void DUdpSocket::resolveFinished(const boost::system::error_code& errorCode,
                                 boost::asio::ip::udp::resolver::iterator iterator)
{
    boost::asio::ip::udp::resolver::iterator endPointIter;//iterates through the list of endpoints
    boost::asio::ip::udp::endpoint endpoint;//the current boost endpoint to examine
    boost::asio::ip::address currentAddress;//the current IP address that was returned
    vector<string> addressList;//a list of the returned IP addresses

    if (errorCode)
    {
        errMsg = "Couldn't resolve hostname: ";
        errMsg.append(errorCode.message());
        if (haveResolveHandler)
            resolveHandler(addressList);
    }//if errorCode

    else
    {
        endPointIter = iterator;

        while (endPointIter != boost::asio::ip::udp::resolver::iterator())
        {
            endpoint = *endPointIter;
            currentAddress = endpoint.address();

            addressList.push_back(currentAddress.to_string());

            endPointIter++;
        }//while endPointIter

        if (haveResolveHandler)
            resolveHandler(addressList);
    }//else errorCode

    //### Allow another hostname resolution to take place ###
    haveResolveHandler = false;
}//resolveFinished

bool DUdpSocket::bind(IpVersion ipVersion,
                      uint16_t portNumber)
{
    boost::system::error_code errorCode;//set if an error occurs in opening or binding
    bool result = true;//false if an error occurred

    //### Close the socket if it's already open ###
    if (!closeSocket())
        result = false;

    //### Open the socket ###
    if (result)
    {
        if (ipVersion == DA_IPV4)
            udpSocket->open(boost::asio::ip::udp::v4(), errorCode);
        else
            udpSocket->open(boost::asio::ip::udp::v6(), errorCode);
        if (errorCode)
        {
            errMsg = "Couldn't open UDP socket. ";
            errMsg.append(errorCode.message());
            result = false;
        }//if errorCode
    }//if result

    //### Bind the socket ###
    if (result)
    {
        if (ipVersion == DA_IPV4)
            udpSocket->bind(boost::asio::ip::udp::endpoint(boost::asio::ip::udp::v4(), portNumber), errorCode);
        else
            udpSocket->bind(boost::asio::ip::udp::endpoint(boost::asio::ip::udp::v6(), portNumber), errorCode);
        if (errorCode)
        {
            errMsg = "Couldn't bind UDP socket. ";
            errMsg.append(errorCode.message());
            result = false;
        }//if errorCode
        else
        {
            socketBound = true;
            myPort = portNumber;
            startReceive();
        }//else errorCode
    }//if result

    return result;
}//bind

bool DUdpSocket::closeSocket(void)
{
    boost::system::error_code errorCode;//set if we failed to close the socket
    bool result = true;//false if we couldn't close the socket

    if (udpSocket->is_open())
    {
        udpSocket->close(errorCode);
        if (errorCode)
        {
            errMsg = "Couldn't close UDP socket. ";
            errMsg.append(errorCode.message());
            result = false;
        }//if errorCode
        else
            myPort = 0;
    }//if is_open()

    socketBound = false;

    return result;
}//closeSocket

void DUdpSocket::send(const libdasyne::DBuffer &theData,
                      const std::string &ipAddress,
                      uint16_t destPort)
{
    DUdpMessage *newMessage = new DUdpMessage(ipAddress, destPort);

    if (socketBound)
    {
        if (theData.size() > 0)
        {
            if (theData.size() > 65535)
            {
                DEngine::addFatal("DUdpSocket::send(): can send at most 65535 bytes at a time, got ");
                DEngine::fatalError(DEngine::intToStr(theData.size()));
            }

            newMessage->setData(theData);
            sendQueue.push_back(newMessage);
            sendNextMessage();
        }
    }
    else
        reportError("DUdpSocket::send(): called, but socket is not bound", false);
}//send

void DUdpSocket::sendNextMessage(void)
{
    DUdpMessage *theMessage = NULL;//contains the UDP payload
    vector<boost::asio::const_buffer> sendBuffers;//the buffers to send
    boost::asio::ip::udp::endpoint destination;//contains the IP address and port number
    DBuffer *payload = NULL;//the data to send
    string destAddress;//the destination address
    uint16_t destPort = 0;//the destination port number

    //### Clear any pending messages if the socket is unbound ###
    if (!socketBound)
    {
        for (unsigned int i = 0; i < sendQueue.size(); i++)
            delete sendQueue.at(i);
        sendQueue.clear();
    }

    if (!isSending)
    {
        if (sendQueue.size() > 0)
        {
            //### Extract the next message ###
            theMessage = sendQueue.at(0);

            //### Obtain the message payload and destination ###
            payload = theMessage->getData();
            destAddress = theMessage->getAddress();
            destPort = theMessage->getPort();

            //### Convert to Boost format ###
            sendBuffers = payload->getSendBuffers(payload->size());
            destination.address(boost::asio::ip::address::from_string(destAddress));
            destination.port(destPort);

            //### Send the message ###
            udpSocket->async_send_to(sendBuffers, destination,
                                     boost::bind(&DUdpSocket::sendFinished, this,
                                                 boost::asio::placeholders::error,
                                                 boost::asio::placeholders::bytes_transferred));
            isSending = true;
        }//if size()

        else if (haveSendHandler)
            sendHandler();
    }//if !isSending
}//sendNextMessage

void DUdpSocket::sendFinished(const boost::system::error_code& errorCode,
                              size_t bytesSent)
{
    //### Verify that we're sending something ###
    if (!isSending)
        netEngine->fatalError("DUdpSocket::sendFinished(): called, but we're not sending!");

    if (bytesSent >= 1)
    {
        delete sendQueue.at(0);
        sendQueue.pop_front();
    }//if bytesSent
    else
    {
        if (udpSocket->is_open())
        {
            errMsg = "Sent zero bytes";
            reportError(errMsg, false);
        }//if is_open()
    }//else bytesSent

    isSending = false;

    if (errorCode)
    {
        if (errorCode != boost::asio::error::operation_aborted)
            recoverableError(errorCode);
    }//if errorCode
    else
        sendNextMessage();
}//sendFinished

void DUdpSocket::startReceive(void)
{
    udpSocket->async_receive_from(boost::asio::buffer(receiveBuffer, RECEIVE_BUFFER_SIZE),
                                  senderEndpoint,
                                  boost::bind(&DUdpSocket::receiveFinished, this,
                                              boost::asio::placeholders::error,
                                              boost::asio::placeholders::bytes_transferred));
}//startReceive

void DUdpSocket::receiveFinished(const boost::system::error_code &errorCode,
                                 size_t bytesReceived)
{
    DBuffer receivedData;//the data received
    string sourceAddress;//the address the message was sent from
    uint16_t sourcePort = 0;//the port number the message was sent from

    if (errorCode)
    {
        if (errorCode != boost::asio::error::operation_aborted)
            recoverableError(errorCode);
    }//if errorCode
    else
    {
        receivedData.appendArray(receiveBuffer, bytesReceived);

        if (receivedData.size() > 0)
        {
            if (haveReceiveHandler)
            {
                sourceAddress = senderEndpoint.address().to_string();
                sourcePort = senderEndpoint.port();

                receiveHandler(receivedData, sourceAddress, sourcePort);
            }//if haveReceiveHandler
        }//if size()

        else
        {
            if (udpSocket->is_open())
            {
                errMsg = "Received zero bytes.";
                reportError(errMsg, false);
            }//if is_open()
        }//else size()

        if (socketBound)
            startReceive();
    }//else errorCode
}//receiveFinished

void DUdpSocket::setReceiveHandler(boost::function<void (DBuffer&, string&, uint16_t)> newHandler)
{
    receiveHandler = newHandler;
    haveReceiveHandler = true;
}//setReceiveHandler

void DUdpSocket::recoverableError(const boost::system::error_code &errorCode)
{
    #ifndef DA_WINDOWS_VERSION
        boost::system::error_condition theError = errorCode.default_error_condition();
    #endif
    string errMsg;//the error to report

    #ifdef DA_WINDOWS_VERSION
        errMsg = errorCode.message();
        errMsg.append(": ");
        errMsg.append(DEngine::intToStr(errorCode.value()));
    #else
        errMsg = theError.message();
        errMsg.append(": ");
        errMsg.append(DEngine::intToStr(theError.value()));
    #endif

    reportError(errMsg, false);
}//recoverableError

void DUdpSocket::reportError(const std::string &errMsg,
                             bool socketOK)
{
    //### Report recoverable errors ###
    if (socketOK)
    {
        if (haveErrorHandler)
            errorHandler(errMsg, socketOK);
    }//if socketOK

    //### Report unrecoverable errors ###
    else
    {
        if (udpSocket->is_open())
        {
            if (haveErrorHandler)
                errorHandler(errMsg, socketOK);
            closeSocket();
        }//if is_open()
    }//else socketOK
}//reportError

void DUdpSocket::setSendHandler(boost::function<void (void)> newHandler)
{
    sendHandler = newHandler;
    haveSendHandler = true;
}//setSendHandler

void DUdpSocket::setErrorHandler(boost::function<void (std::string, bool)> newHandler)
{
    errorHandler = newHandler;
    haveErrorHandler = true;
}//setErrorHandler

void DUdpSocket::clearAllHandlers(void)
{
    haveSendHandler = false;
    haveErrorHandler = false;
    haveReceiveHandler = false;
    haveResolveHandler = false;
}//clearAllHandlers

uint16_t DUdpSocket::getPortNumber(void)
{
    return myPort;
}//getPortNumber

string DUdpSocket::getErrMsg(void)
{
    return errMsg;
}//getErrMsg
