/*
 * Copyright (C) 2014-2015 David C. McCallum
 * For conditions of distribution and use, see the copyright notice in Dasyne.hpp
*/

/* DESCRIPTION
    Wrapper around a boost::ASIO udp socket.
*/

#ifndef DUDPSOCKET_HPP_INCLUDED
#define DUDPSOCKET_HPP_INCLUDED

#include "DEngine.hpp"
#include "DSocket.hpp"
#include <deque>

namespace libdasyne
{
    class DBuffer;
    class DUdpMessage;

    class DUdpSocket : public DSocket
    {
        public:
            //constructor
            DUdpSocket(DEngine *newNetEngine);

            //destructor
            virtual ~DUdpSocket(void);

            //Given a host name of the form www.google.com, returns the associated IP address
            //in the specified event handler. Uses DNS.
            void resolveHost(const std::string &hostName,
                             libdasyne::IpVersion ipVersion,
                             boost::function<void (std::vector<std::string>)> newResolveHandler);

            //Bind the UDP socket to the specified port number. Returns false if an error occurred
            //and sets errMsg.
            bool bind(IpVersion ipVersion,
                      uint16_t portNumber);

            //Close a socket that was opened with bind(). Returns false on failure and sets errMsg
            //with a description of the problem.
            virtual bool closeSocket(void);

            //Sends theData to the specified IP address and port number.
            void send(const libdasyne::DBuffer &theData,
                      const std::string &ipAddress,
                      uint16_t destPort);

            //Event handler, called when there are no more bytes to send
            void setSendHandler(boost::function<void (void)> newHandler);

            //Event handler, called when an error occurs. If socketOK is false, the socket was closed
            void setErrorHandler(boost::function<void (std::string, bool)> newHandler);

            //Event handler, called when we've received data
            void setReceiveHandler(boost::function<void (DBuffer&, std::string&, uint16_t)> newHandler);

            //Returns a description of the most recent error with this socket
            std::string getErrMsg(void);

            //Returns the port to which this socket is bound, or 0 if none
            uint16_t getPortNumber(void);

            //Removes all user-defined event handlers from this socket.
            void clearAllHandlers(void);

            //The number of bytes in our receive buffer, this is the maximum payload size that we can receive.
            static const unsigned int RECEIVE_BUFFER_SIZE = 65535;

        protected:
            //Calls errorHandler and closes the socket if socketOK is false
            void reportError(const std::string &errMsg,
                             bool socketOK);

        private:
            //Data that's waiting to be sent to the remote host
            std::deque<DUdpMessage*> sendQueue;

            //Handles events for this object
            DEngine *netEngine;

            //An array of bytes to hold received data
            unsigned char *receiveBuffer;

            //The UDP socket itself
            boost::asio::ip::udp::socket *udpSocket;

            //resolves host names into a boost endpoint
            boost::asio::ip::udp::resolver *hostResolver;

            //Event handler to call if an error occurs.
            boost::function<void (std::string, bool)> errorHandler;

            //Event handler that's called when we've finished sending all data. This variable is only
            //valid if haveSendHandler is true.
            boost::function<void (void)> sendHandler;

            //Event handler called when resolveHost finishes
            boost::function<void (std::vector<std::string>)> resolveHandler;

            //Event handler that's called when we receive data
            boost::function<void (DBuffer&, std::string&, uint16_t)> receiveHandler;

            //The port number and IP address of the remote host that sent the currently received packet
            boost::asio::ip::udp::socket::basic_datagram_socket::endpoint_type senderEndpoint;

            //A description of the most recent error
            std::string errMsg;

            //The port to which this socket is bound, or 0 for none
            uint16_t myPort;

            //True if we have the corresponding event handler
            bool haveSendHandler;
            bool haveErrorHandler;
            bool haveReceiveHandler;
            bool haveResolveHandler;

            //True if we're currently sending a packet, false if the socket is idle
            bool isSending;

            //True if the socket is currently bound and ready for use, false otherwise.
            bool socketBound;

            //############### FUNCTIONS ###############

            //Sends the next message in the sendQueue
            void sendNextMessage(void);

            //Event handler, called when an asynchonous send operation completes
            void sendFinished(const boost::system::error_code& errorCode,
                              size_t bytesSent);

            //Calls reportError with a description of the problem, with socketOK set to true
            void recoverableError(const boost::system::error_code &errorCode);

            //Start an asynchronous receive operation
            void startReceive(void);

            //Event handler, called when we've finished receiving data
            void receiveFinished(const boost::system::error_code &errorCode,
                                 size_t bytesReceived);

            //Event handler, called when host name resolution is complete
            void resolveFinished(const boost::system::error_code& error,
                                 boost::asio::ip::udp::resolver::iterator iterator);

    };//DUdpSocket
}//namespace libdasyne

#endif
